# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# type: ignore

import math

import numpy as np

from mava.utils.environments.RoboCup_env.robocup_utils.game_object import Flag
from mava.utils.environments.RoboCup_env.robocup_utils.util_functions import (
    rad_rot_to_xy,
)

true_flag_coords = Flag.FLAG_COORDS


def rotate(origin, point, angle):
    """
    Rotate a point counterclockwise by a given angle around a given origin.

    The angle should be given in radians.
    """

    ox, oy = origin
    px, py = point

    qx = ox + math.cos(angle) * (px - ox) - math.sin(angle) * (py - oy)
    qy = oy + math.sin(angle) * (px - ox) + math.cos(angle) * (py - oy)
    return qx, qy


def rel_to_abs_coords(obj):
    obj_x = obj.distance * math.sin(-obj.direction * math.pi / 180)
    obj_y = -obj.distance * math.cos(-obj.direction * math.pi / 180)
    return obj_x, obj_y


class WorldModel:
    """
    Holds and updates the model of the world as known from current and past
    data.
    """

    # constants for team sides
    SIDE_L = "l"
    SIDE_R = "r"

    class PlayModes:
        """
        Acts as a static class containing variables for all valid play modes.
        The string values correspond to what the referee calls the game modes.
        """

        BEFORE_KICK_OFF = "before_kick_off"
        PLAY_ON = "play_on"
        TIME_OVER = "time_over"
        KICK_OFF_L = "kick_off_l"
        KICK_OFF_R = "kick_off_r"
        KICK_IN_L = "kick_in_l"
        KICK_IN_R = "kick_in_r"
        FREE_KICK_L = "free_kick_l"
        FREE_KICK_R = "free_kick_r"
        CORNER_KICK_L = "corner_kick_l"
        CORNER_KICK_R = "corner_kick_r"
        GOAL_KICK_L = "goal_kick_l"
        GOAL_KICK_R = "goal_kick_r"
        DROP_BALL = "drop_ball"
        OFFSIDE_L = "offside_l"
        OFFSIDE_R = "offside_r"

        def __init__(self):
            raise NotImplementedError(
                "Don't instantiate a PlayModes class,"
                " access it statically through WorldModel instead."
            )

    class RefereeMessages:
        """
        Static class containing possible non-mode messages sent by a referee.
        """

        # these are referee messages, not play modes
        FOUL_L = "foul_l"
        FOUL_R = "foul_r"
        GOALIE_CATCH_BALL_L = "goalie_catch_ball_l"
        GOALIE_CATCH_BALL_R = "goalie_catch_ball_r"
        TIME_UP_WITHOUT_A_TEAM = "time_up_without_a_team"
        TIME_UP = "time_up"
        HALF_TIME = "half_time"
        TIME_EXTENDED = "time_extended"

        # these are special, as they are always followed by '_' and an int of
        # the number of goals scored by that side so far.  these won't match
        # anything specifically, but goals WILL start with these.
        GOAL_L = "goal_l_"
        GOAL_R = "goal_r_"

        def __init__(self):
            raise NotImplementedError(
                "Don't instantiate a RefereeMessages class,"
                " access it statically through WorldModel instead."
            )

    def __init__(self, action_handler):
        """
        Create the world model with default values and an ActionHandler class it
        can use to complete requested actions.
        """
        # we use the action handler to complete complex commands
        self.ah = action_handler

        # these variables store all objects for any particular game step
        self.ball = None
        self.flags = []
        self.goals = []
        self.players = []
        self.lines = []

        # the default position of this player, its home position
        self.home_point = (None, None)

        # scores for each side
        self.score_l = 0
        self.score_r = 0

        # the name of the agent's team
        self.teamname = None

        # handle player information, like uniform number and side
        self.side = None
        self.uniform_number = None

        # stores the most recent message heard
        self.last_message = None

        # the mode the game is currently in (default to not playing yet)
        self.play_mode = WorldModel.PlayModes.BEFORE_KICK_OFF

        # Obs updated
        # self.obs_updated = False

        # body state
        self.view_width = None
        self.view_quality = None
        self.stamina = None
        self.effort = None
        self.speed_amount = None
        self.speed_direction = None
        self.neck_direction = None
        self.new_data = False

        # counts of actions taken so far
        self.kick_count = None
        self.dash_count = None
        self.turn_count = None
        self.say_count = None
        self.turn_neck_count = None
        self.catch_count = None
        self.move_count = None
        self.change_view_count = None

        # apparent absolute player coordinates and neck/body directions
        self.abs_coords = (None, None)
        self.abs_body_dir = None
        self.abs_neck_dir = None

        # create a new server parameter object for holding all server params
        self.server_parameters = ServerParameters()

    def __calculate_abs_info(self):
        if self.flags is not None:
            rel_coords = []
            true_coords = []
            for flag in self.flags:
                if (
                    flag is not None
                    and flag.direction is not None
                    and flag.distance is not None
                    and flag.flag_id is not None
                ):
                    obs_coords = rel_to_abs_coords(flag)
                    rel_coords.append(obs_coords)
                    true_coords.append(true_flag_coords[flag.flag_id])

            for goal in self.goals:
                if (
                    goal is not None
                    and goal.direction is not None
                    and goal.distance is not None
                    and goal.goal_id is not None
                ):
                    obs_coords = rel_to_abs_coords(goal)
                    rel_coords.append(obs_coords)
                    true_coords.append(true_flag_coords[goal.goal_id])

            if len(true_coords) > 1 and len(rel_coords) > 1:
                # Get means
                rel_mean = np.mean(rel_coords, axis=0)
                true_mean = np.mean(true_coords, axis=0)

                mean_off = rel_mean - true_mean

                # Get rotation
                rel_de_mean = np.array(rel_coords) - rel_mean
                true_de_mean = np.array(true_coords) - true_mean

                if len(true_de_mean) > 1:
                    ang_offs = np.arctan2(
                        true_de_mean[:, 1], true_de_mean[:, 0]
                    ) - np.arctan2(rel_de_mean[:, 1], rel_de_mean[:, 0])
                    x, y = rad_rot_to_xy(ang_offs)

                    x_mean = np.mean(x)
                    y_mean = np.mean(y)

                    ang_offs = np.arctan2(y_mean, x_mean)

                    true_agent_loc = rotate(rel_mean, (0, 0), ang_offs) - mean_off
                    self.abs_coords = true_agent_loc
                    self.abs_body_dir = (ang_offs / math.pi) * 180

                    # TODO: Is this correct?
                    self.abs_neck_dir = self.abs_body_dir + self.neck_direction

    def process_new_info(self, ball, flags, goals, players, lines):
        """
        Update any internal variables based on the currently available
        information.  This also calculates information not available directly
        from server-reported messages, such as player coordinates.
        """

        # update basic information
        self.ball = ball
        self.flags = flags
        self.goals = goals
        self.players = players
        self.lines = lines

        self.__calculate_abs_info()

    def is_playon(self):
        """
        Tells us whether it's play time
        """
        return (
            self.play_mode == WorldModel.PlayModes.PLAY_ON
            or self.play_mode == WorldModel.PlayModes.KICK_OFF_L
            or self.play_mode == WorldModel.PlayModes.KICK_OFF_R
            or self.play_mode == WorldModel.PlayModes.KICK_IN_L
            or self.play_mode == WorldModel.PlayModes.KICK_IN_R
            or self.play_mode == WorldModel.PlayModes.FREE_KICK_L
            or self.play_mode == WorldModel.PlayModes.FREE_KICK_R
            or self.play_mode == WorldModel.PlayModes.CORNER_KICK_L
            or self.play_mode == WorldModel.PlayModes.CORNER_KICK_R
            or self.play_mode == WorldModel.PlayModes.GOAL_KICK_L
            or self.play_mode == WorldModel.PlayModes.GOAL_KICK_R
            or self.play_mode == WorldModel.PlayModes.DROP_BALL
            or self.play_mode == WorldModel.PlayModes.OFFSIDE_L
            or self.play_mode == WorldModel.PlayModes.OFFSIDE_R
        )

    def is_before_kick_off(self):
        """
        Tells us whether the game is in a pre-kickoff state.
        """

        return self.play_mode == WorldModel.PlayModes.BEFORE_KICK_OFF

    def is_kick_off_us(self):
        """
        Tells us whether it's our turn to kick off.
        """

        ko_left = WorldModel.PlayModes.KICK_OFF_L
        ko_right = WorldModel.PlayModes.KICK_OFF_R

        # return whether we're on the side that's kicking off
        return (
            self.side == WorldModel.SIDE_L
            and self.play_mode == ko_left
            or self.side == WorldModel.SIDE_R
            and self.play_mode == ko_right
        )

    def is_dead_ball_them(self):
        """
        Returns whether the ball is in the other team's posession and it's a
        free kick, corner kick, or kick in.
        """

        # shorthand for verbose constants
        kil = WorldModel.PlayModes.KICK_IN_L
        kir = WorldModel.PlayModes.KICK_IN_R
        fkl = WorldModel.PlayModes.FREE_KICK_L
        fkr = WorldModel.PlayModes.FREE_KICK_R
        ckl = WorldModel.PlayModes.CORNER_KICK_L
        ckr = WorldModel.PlayModes.CORNER_KICK_R

        # shorthand for whether left team or right team is free to act
        pm = self.play_mode
        free_left = pm == kil or pm == fkl or pm == ckl
        free_right = pm == kir or pm == fkr or pm == ckr

        # return whether the opposing side is in a dead ball situation
        if self.side == WorldModel.SIDE_L:
            return free_right
        else:
            return free_left

    def is_ball_kickable(self):
        """
        Tells us whether the ball is in reach of the current player.
        """

        # ball must be visible, not behind us, and within the kickable margin
        return (
            self.ball is not None
            and self.ball.distance is not None
            and self.ball.distance <= self.server_parameters.kickable_margin
        )

    def get_ball_speed_max(self):
        """
        Returns the maximum speed the ball can be kicked at.
        """

        return self.server_parameters.ball_speed_max

    def get_stamina(self):
        """
        Returns the agent's current stamina amount.
        """

        return self.stamina

    def get_stamina_max(self):
        """
        Returns the maximum amount of stamina a player can have.
        """

        return self.server_parameters.stamina_max

    def turn_body_to_object(self, obj):
        """
        Turns the player's body to face a particular object.
        """

        self.ah.turn(obj.direction)


class ServerParameters:
    """
    A storage container for all the settings of the soccer server.
    """

    def __init__(self):
        """
        Initialize default parameters for a server.
        """

        self.audio_cut_dist = 50
        self.auto_mode = 0
        self.back_passes = 1
        self.ball_accel_max = 2.7
        self.ball_decay = 0.94
        self.ball_rand = 0.05
        self.ball_size = 0.085
        self.ball_speed_max = 2.7
        self.ball_stuck_area = 3
        self.ball_weight = 0.2
        self.catch_ban_cycle = 5
        self.catch_probability = 1
        self.catchable_area_l = 2
        self.catchable_area_w = 1
        self.ckick_margin = 1
        self.clang_advice_win = 1
        self.clang_define_win = 1
        self.clang_del_win = 1
        self.clang_info_win = 1
        self.clang_mess_delay = 50
        self.clang_mess_per_cycle = 1
        self.clang_meta_win = 1
        self.clang_rule_win = 1
        self.clang_win_size = 300
        self.coach = 0
        self.coach_port = 6001
        self.coach_w_referee = 0
        self.connect_wait = 300
        self.control_radius = 2
        self.dash_power_rate = 0.006
        self.drop_ball_time = 200
        self.effort_dec = 0.005
        self.effort_dec_thr = 0.3
        self.effort_inc = 0.01
        self.effort_inc_thr = 0.6
        self.effort_init = 1
        self.effort_min = 0.6
        self.forbid_kick_off_offside = 1
        self.free_kick_faults = 1
        self.freeform_send_period = 20
        self.freeform_wait_period = 600
        self.fullstate_l = 0
        self.fullstate_r = 0
        self.game_log_compression = 0
        self.game_log_dated = 1
        self.game_log_dir = "./"
        self.game_log_fixed = 0
        self.game_log_fixed_name = "rcssserver"
        self.game_log_version = 3
        self.game_logging = 1
        self.game_over_wait = 100
        self.goal_width = 14.02
        self.goalie_max_moves = 2
        self.half_time = 300
        self.hear_decay = 1
        self.hear_inc = 1
        self.hear_max = 1
        self.inertia_moment = 5
        self.keepaway = 0
        self.keepaway_length = 20
        self.keepaway_log_dated = 1
        self.keepaway_log_dir = "./"
        self.keepaway_log_fixed = 0
        self.keepaway_log_fixed_name = "rcssserver"
        self.keepaway_logging = 1
        self.keepaway_start = -1
        self.keepaway_width = 20
        self.kick_off_wait = 100
        self.kick_power_rate = 0.027
        self.kick_rand = 0
        self.kick_rand_factor_l = 1
        self.kick_rand_factor_r = 1
        self.kickable_margin = 0.7
        self.landmark_file = "~/.rcssserver-landmark.xml"
        self.log_date_format = "%Y%m%d%H%M-"
        self.log_times = 0
        self.max_goal_kicks = 3
        self.maxmoment = 180
        self.maxneckang = 90
        self.maxneckmoment = 180
        self.maxpower = 100
        self.minmoment = -180
        self.minneckang = -90
        self.minneckmoment = -180
        self.minpower = -100
        self.nr_extra_halfs = 2
        self.nr_normal_halfs = 2
        self.offside_active_area_size = 2.5
        self.offside_kick_margin = 9.15
        self.olcoach_port = 6002
        self.old_coach_hear = 0
        self.pen_allow_mult_kicks = 1
        self.pen_before_setup_wait = 30
        self.pen_coach_moves_players = 1
        self.pen_dist_x = 42.5
        self.pen_max_extra_kicks = 10
        self.pen_max_goalie_dist_x = 14
        self.pen_nr_kicks = 5
        self.pen_random_winner = 0
        self.pen_ready_wait = 50
        self.pen_setup_wait = 100
        self.pen_taken_wait = 200
        self.penalty_shoot_outs = 1
        self.player_accel_max = 1
        self.player_decay = 0.4
        self.player_rand = 0.1
        self.player_size = 0.3
        self.player_speed_max = 1.2
        self.player_weight = 60
        self.point_to_ban = 5
        self.point_to_duration = 20
        self.port = 6000
        self.prand_factor_l = 1
        self.prand_factor_r = 1
        self.profile = 0
        self.proper_goal_kicks = 0
        self.quantize_step = 0.1
        self.quantize_step_l = 0.01
        self.record_messages = 0
        self.recover_dec = 0.002
        self.recover_dec_thr = 0.3
        self.recover_init = 1
        self.recover_min = 0.5
        self.recv_step = 10
        self.say_coach_cnt_max = 128
        self.say_coach_msg_size = 128
        self.say_msg_size = 10
        self.send_comms = 0
        self.send_step = 150
        self.send_vi_step = 100
        self.sense_body_step = 100
        self.simulator_step = 100
        self.slow_down_factor = 1
        self.slowness_on_top_for_left_team = 1
        self.slowness_on_top_for_right_team = 1
        self.stamina_inc_max = 45
        self.stamina_max = 4000
        self.start_goal_l = 0
        self.start_goal_r = 0
        self.stopped_ball_vel = 0.01
        self.synch_micro_sleep = 1
        self.synch_mode = 0
        self.synch_offset = 60
        self.tackle_back_dist = 0.5
        self.tackle_cycles = 10
        self.tackle_dist = 2
        self.tackle_exponent = 6
        self.tackle_power_rate = 0.027
        self.tackle_width = 1
        self.team_actuator_noise = 0
        self.text_log_compression = 0
        self.text_log_dated = 1
        self.text_log_dir = "./"
        self.text_log_fixed = 0
        self.text_log_fixed_name = "rcssserver"
        self.text_logging = 1
        self.use_offside = 1
        self.verbose = 0
        self.visible_angle = 90
        self.visible_distance = 3
        self.wind_ang = 0
        self.wind_dir = 0
        self.wind_force = 0
        self.wind_none = 0
        self.wind_rand = 0
        self.wind_random = 0



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

import sonnet as snt

from mava.components.tf.architectures import BaseArchitecture
from mava.components.tf.modules.mixing.base import BaseMixingModule
from mava.components.tf.networks.additive import AdditiveMixingNetwork


class AdditiveMixing(BaseMixingModule):
    """Multi-agent monotonic mixing architecture."""

    def __init__(self, architecture: BaseArchitecture) -> None:
        """Initializes the mixer."""
        super(AdditiveMixing, self).__init__()

        self._architecture = architecture
        self._agent_networks = self._architecture.create_actor_variables()

    def _create_mixing_layer(self) -> snt.Module:
        # Instantiate additive mixing network
        self._mixed_network = AdditiveMixingNetwork()
        return self._mixed_network

    def create_system(self) -> Dict[str, Dict[str, snt.Module]]:
        self._agent_networks["mixing"] = self._create_mixing_layer()
        self._agent_networks["target_mixing"] = self._create_mixing_layer()

        return self._agent_networks



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# type: ignore

import re

# used to convert server value strings into actual python values
pattern_int = re.compile(r"^-?\d+$")
pattern_float = re.compile(r"^-?\d*[.]\d+$")


def parse(text):  # noqa: C901
    """
    This is what amounts to a simple lisp parser for turning the server's
    returned messages into an intermediate format that's easier to deal
    with than the raw (often poorly formatted) text.

    This parses generally, taking any lisp-like string and turning it into a
    list of nested lists, where each nesting indicates a parenthesized
    expression.  holding multiple top-level parenthesized expressions. Ex: "(baz
    0 (foo 1.5))" becomes ['baz', 0, ['foo', 1.5]].
    """

    text = text.decode()  # Added for python 3

    # make sure all of our parenthesis match
    if text.count("(") == text.count(")"):
        # result acts as a stack that holds the strings grouped by nested parens.
        # result will only ever contain one item, the first level of indenting
        # encountered.  this is because the server (hopefully!) only ever sends one
        # message at a time.
        # TODO: make sure that the server only ever sends one message at a time!
        result = []

        # the current level of indentation, used to append chars to correct level
        indent = 0

        # the non-indenting characters we find. these are kept in a buffer until
        # we indent or dedent, and then are added to the current indent level all
        # at once, for efficiency.
        s = []

        # whether we're currently in the middle of parsing a string
        in_string = False

        # the last character seen, None to begin with
        prev_c = None

        for c in text:
            # prevent parsing parens when inside a string (also ignores escaped
            # '"'s as well). doesn't add the quotes so we don't have to recognize
            # that value as a string via a regex.
            if c == '"' and prev_c != "\\":
                in_string = not in_string

            # we only indent/dedent if not in the middle of parsing a string
            elif c == "(" and not in_string:
                # recurse into current level of nesting
                cur = result
                for i in range(indent):
                    cur = cur[-1]

                # add our buffered string onto the previous level, then clear it
                # for the next.
                if len(s) > 0:
                    val = "".join(s)

                    # try to convert our string into a value and append it to our
                    # list.  failing that, simply append it as an attribute name.
                    if pattern_int.match(val):
                        cur.append(int(val))
                    elif pattern_float.match(val):
                        cur.append(float(val))
                    else:
                        cur.append(val)

                    s = []

                # append a new level of nesting to our list
                cur.append([])

                # increase the indent level so we can get back to this level later
                indent += 1

            elif c == ")" and not in_string:
                # append remaining string buffer before dedenting
                if len(s) > 0:
                    cur = result
                    for i in range(indent):
                        cur = cur[-1]

                    val = "".join(s)

                    # try to convert our string into a value and append it to our
                    # list.  failing that, simply append it as an attribute name.
                    if pattern_int.match(val):
                        cur.append(int(val))
                    elif pattern_float.match(val):
                        cur.append(float(val))
                    else:
                        cur.append(val)

                    s = []

                # we finished with one level, so dedent back to the previous one
                indent -= 1

            # append non-space characters to the buffer list. spaces are delimiters
            # for expressions, hence are special.
            elif c != " ":
                # append the current string character to the buffer list.
                s.append(c)

            # we separate expressions by spaces
            elif c == " " and len(s) > 0:
                cur = result
                for i in range(indent):
                    cur = cur[-1]

                val = "".join(s)

                # try to convert our string into a value and append it to our
                # list.  failing that, simply append it as an attribute name.
                if pattern_int.match(val):
                    cur.append(int(val))
                elif pattern_float.match(val):
                    cur.append(float(val))
                else:
                    cur.append(val)

                s = []

            # save the previous character. used to determine if c is escaped
            prev_c = c

        # this returns the first and only message found.  result is a list simply
        # because it makes adding new levels of indentation simpler as it avoids
        # the 'if result is None' corner case that would come up when trying to
        # append the first '('.
        return result[0]
    else:
        return None



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# type: ignore

import collections
import queue

from mava.utils.environments.RoboCup_env.robocup_utils import (
    game_object,
    message_parser,
    sp_exceptions,
)
from mava.utils.environments.RoboCup_env.robocup_utils.player_world_model import (
    WorldModel,
)

# should we print messages received from the server?
PRINT_SERVER_MESSAGES = False

# should we print commands sent to the server?
PRINT_SENT_COMMANDS = False


class MessageHandler:
    """
    Handles all incoming messages from the server.  Parses their data and puts
    it into the given WorldModel.

    All '_handle_*' functions deal with their appropriate message types
    as received from a server.  This allows adding a message handler to be as
    simple as adding a new '_handle_*' function to this object.
    """

    # an inner class used for creating named tuple 'hear' messages
    Message = collections.namedtuple("Message", "time sender message")

    def __init__(self, world_model):
        self.wm = world_model

    def handle_message(self, msg):
        """
        Takes a raw message direct from the server, parses it, and stores its
        data in the world and body model objects given at init.  Returns the
        type of message received.
        """

        # get all the expressions contained in the given message
        parsed = message_parser.parse(msg)

        if parsed is not None:
            if PRINT_SERVER_MESSAGES:
                print(parsed[0] + ":", parsed[1:], "\n")

            # this is the name of the function that should be used to handle
            # this message type.  we pull it from this object dynamically to
            # avoid having a huge if/elif/.../else statement.

            if parsed[0] != "ok" and parsed[0] != "think":
                msg_func = "_handle_%s" % parsed[0]

                if hasattr(self, msg_func):
                    # call the appropriate function with this message
                    getattr(self, msg_func).__call__(parsed)

                # throw an exception if we don't know about the given message type
                else:
                    m = "Can't handle message type '%s', function '%s' not found."
                    raise sp_exceptions.MessageTypeError(m % (parsed[0], msg_func))

                # return the type of message received
                return parsed[0]
        else:
            return parsed

    def _handle_see(self, msg):  # noqa: C901
        """
        Parses visual information in a message and turns it into useful data.

        This comes to us as a list of lists.  Each list contains another list as
        its first element, the contents of which describe a particular object.
        The other items of the list are data pertaining to the object.  We parse
        each list into its own game object, then insert those game objects into
        the world model.
        """

        # the simulation cycle of the soccer server
        # TODO: we should probably use this somewhere...
        # sim_time = msg[1]

        # store new values before changing those in the world model.  all new
        # values replace those in the world model at the end of parsing.
        new_ball = None
        new_flags = []
        new_goals = []
        new_lines = []
        new_players = []

        # iterate over all the objects given to us in the last see message
        for obj in msg[2:]:
            name = obj[0]
            members = obj[1:]

            # get basic information from the object.  different numbers of
            # parameters (inconveniently) specify different types and
            # arrangements of data received for the object.

            # default values for object data
            distance = None
            direction = None
            dist_change = None
            dir_change = None
            body_dir = None
            neck_dir = None

            # a single item object means only direction
            if len(members) == 1:
                direction = members[0]

            # objects with more items follow a regular pattern
            elif len(members) >= 2:
                distance = members[0]
                direction = members[1]

                # include delta values if present
                if len(members) >= 4:
                    dist_change = members[2]
                    dir_change = members[3]

                # include body/neck values if present
                if len(members) >= 6:
                    body_dir = members[4]
                    neck_dir = members[5]

            # parse flags
            if name[0] == "f":
                # since the flag's name sometimes contains a number, the parser
                # recognizes it as such and converts it into an int.  it's
                # always the last item when it's a number, so we stringify the
                # last item of the name to convert any numbers back.
                name[-1] = str(name[-1])

                # the flag's id is its name's members following the f as a string
                flag_id = "".join(name[1:])

                new_flags.append(game_object.Flag(distance, direction, flag_id))

            # parse players
            elif name[0] == "p":
                # extract any available information from the player object's name
                teamname = None
                uniform_number = None

                if len(name) >= 2:
                    teamname = name[1]
                if len(name) >= 3:
                    uniform_number = name[2]
                # if len(name) >= 4:
                #     position = name[3]

                # figure out the player's side
                side = None
                if teamname is not None:
                    # if they're on our team, they're on our side
                    if teamname == self.wm.teamname:
                        side = self.wm.side
                    # otherwise, set side to the other team's side
                    else:
                        if self.wm.side == WorldModel.SIDE_L:
                            side = WorldModel.SIDE_R
                        else:
                            side = WorldModel.SIDE_L

                # calculate player's speed
                speed = None
                # TODO: calculate player's speed!

                new_players.append(
                    game_object.Player(
                        distance,
                        direction,
                        dist_change,
                        dir_change,
                        speed,
                        teamname,
                        side,
                        uniform_number,
                        body_dir,
                        neck_dir,
                    )
                )

            # parse goals
            elif name[0] == "g":
                # see if we know which side's goal this is
                goal_id = None
                if len(name) > 1:
                    goal_id = name[1]

                new_goals.append(game_object.Goal(distance, direction, goal_id))

            # parse lines
            elif name[0] == "l":
                # see if we know which line this is
                line_id = None
                if len(name) > 1:
                    line_id = name[1]

                new_lines.append(game_object.Line(distance, direction, line_id))

            # parse the ball
            elif name[0] == "b":
                # TODO: handle speed!
                new_ball = game_object.Ball(
                    distance, direction, dist_change, dir_change, None
                )

            # object very near to but not viewable by the player are 'blank'

            # the out-of-view ball
            elif name[0] == "B":
                new_ball = game_object.Ball(None, None, None, None, None)

            # an out-of-view flag
            elif name[0] == "F":
                new_flags.append(game_object.Flag(None, None, None))

            # an out-of-view goal
            elif name[0] == "G":
                new_goals.append(game_object.Goal(None, None, None))

            # an out-of-view player
            elif name[0] == "P":
                new_players.append(
                    game_object.Player(
                        None, None, None, None, None, None, None, None, None, None
                    )
                )

            # an unhandled object type
            else:
                raise ValueError("Unknown object: '" + str(obj) + "'")

        # tell the WorldModel to update any internal variables based on the
        # newly gleaned information.
        self.wm.process_new_info(new_ball, new_flags, new_goals, new_players, new_lines)

    def _handle_see_global(self, msg):
        """
        Parses global visual information in a message and turns it into useful data.

        This comes to us as a list of lists.  Each list contains another list as
        its first element, the contents of which describe a particular object.
        The other items of the list are data pertaining to the object.  We parse
        each list into its own game object, then insert those game objects into
        the world model.
        """

        # store new values before changing those in the world model.  all new
        # values replace those in the world model at the end of parsing.
        ball = None
        goals = []
        players = []

        # iterate over all the objects given to us in the last see message
        for obj in msg[2:]:
            name = obj[0]
            info = obj[1:]

            # parse players
            if name[0] == "p":
                assert len(name) == 3
                teamname = name[1]
                uniform_number = name[2]

                if len(info) == 6:
                    x, y, delta_x, delta_y, body_angle, neck_angle = info
                elif len(info) == 7:
                    # TODO: Figure out what this extra is
                    x, y, delta_x, delta_y, body_angle, neck_angle, extra = info
                    assert extra == "t"
                else:
                    raise NotImplementedError("Info size to large: ", len(info))

                players.append(
                    {
                        "teamname": teamname,
                        "uniform_number": uniform_number,
                        "coords": (x, y),
                        "delta_coords": (delta_x, delta_y),
                        "body_angle": body_angle,
                        "neck_angle": neck_angle,
                    }
                )

            # parse goals
            elif name[0] == "g":
                assert len(name) == 2
                assert len(info) == 2

                goal_side = name[1]
                x, y = info

                goals.append({"side": goal_side, "coords": (x, y)})

            # parse the ball
            elif name[0] == "b":
                assert len(name) == 1
                assert len(info) == 4

                x, y, delta_x, delta_y = info
                ball = {"coords": (x, y), "delta_coords": (delta_x, delta_y)}
            else:
                raise ValueError("Unknown object: '" + str(obj) + "'")

        # tell the WorldModel to update any internal variables based on the
        # newly gleaned information.
        self.wm.process_new_info(ball, goals, players)
        self.wm.new_data = True

    def _handle_hear(self, msg):
        """
        Parses audible information and turns it into useful information.
        """

        if str(msg[1]).isdigit():
            # Player
            time_recvd = msg[1]  # server cycle when message was heard
            sender = msg[2]  # name (or direction) of who sent the message
        else:
            # Trainer
            time_recvd = msg[2]  # server cycle when message was heard
            sender = msg[1]  # name (or direction) of who sent the message

        message = msg[3]  # message string

        # ignore messages sent by self (NOTE: would anybody really want these?)
        if sender == "self":
            return

        # handle messages from the referee, to update game state
        elif sender == "referee":
            # change the name for convenience's sake
            mode = message

            # deal first with messages that shouldn't be passed on to the agent

            # keep track of scores by setting them to the value reported.  this
            # precludes any possibility of getting out of sync with the server.
            if mode.startswith(WorldModel.RefereeMessages.GOAL_L):
                # split off the number, the part after the rightmost '_'
                self.wm.score_l = int(mode.rsplit("_", 1)[1])
                return
            elif mode.startswith(WorldModel.RefereeMessages.GOAL_R):
                self.wm.score_r = int(mode.rsplit("_", 1)[1])
                return

            # ignore these messages, but pass them on to the agent. these don't
            # change state but could still be useful.
            elif (
                mode == WorldModel.RefereeMessages.FOUL_L
                or mode == WorldModel.RefereeMessages.FOUL_R
                or mode == WorldModel.RefereeMessages.GOALIE_CATCH_BALL_L
                or mode == WorldModel.RefereeMessages.GOALIE_CATCH_BALL_R
                or mode == WorldModel.RefereeMessages.TIME_UP_WITHOUT_A_TEAM
                or mode == WorldModel.RefereeMessages.HALF_TIME
                or mode == WorldModel.RefereeMessages.TIME_EXTENDED
            ):

                # messages are named 3-tuples of (time, sender, message)
                ref_msg = self.Message(time_recvd, sender, message)

                # pass this message on to the player and return
                self.wm.last_message = ref_msg
                return

            # deal with messages that indicate game mode, but that the agent
            # doesn't need to know about specifically.
            else:
                # set the mode to the referee reported mode string
                self.wm.play_mode = mode
                return

        # all other messages are treated equally
        else:
            # update the model's last heard message
            new_msg = MessageHandler.Message(time_recvd, sender, message)
            self.wm.prev_message = new_msg

    def _handle_sense_body(self, msg):
        """
        Deals with the agent's body model information.
        """

        # update the body model information when received. each piece of info is
        # a list with the first item as the name of the data, and the rest as
        # the values.
        for info in msg[2:]:
            name = info[0]
            values = info[1:]

            if name == "view_mode":
                self.wm.view_quality = values[0]
                self.wm.view_width = values[1]
            elif name == "stamina":
                self.wm.stamina = values[0]
                self.wm.effort = values[1]
            elif name == "speed":
                self.wm.speed_amount = values[0]
                self.wm.speed_direction = values[1]
            elif name == "head_angle":
                self.wm.neck_direction = values[0]

            # these update the counts of the basic actions taken
            elif name == "kick":
                self.wm.kick_count = values[0]
            elif name == "dash":
                self.wm.dash_count = values[0]
            elif name == "turn":
                self.wm.turn_count = values[0]
            elif name == "say":
                self.wm.say_count = values[0]
            elif name == "turn_neck":
                self.wm.turn_neck_count = values[0]
            elif name == "catch":
                self.wm.catch_count = values[0]
            elif name == "move":
                self.wm.move_count = values[0]
            elif name == "change_view":
                self.wm.change_view_count = values[0]

            # we leave unknown values out of the equation
            else:
                pass
        self.wm.new_data = True

    def _handle_change_player_type(self, msg):
        """
        Handle player change messages.
        """

    def _handle_player_param(self, msg):
        """
        Deals with player parameter information.
        """

    def _handle_player_type(self, msg):
        """
        Handles player type information.
        """

    def _handle_server_param(self, msg):
        """
        Stores server parameter information.
        """

        # each list is two items: a value name and its value.  we add them all
        # to the ServerParameters class inside WorldModel programmatically.
        for param in msg[1:]:
            # put all [param, value] pairs into the server settings object
            # by setting the attribute programmatically.
            if len(param) != 2:
                continue

            # the parameter and its value
            key = param[0]
            value = param[1]

            # set the attribute if it was accounted for, otherwise alert the user
            if hasattr(self.wm.server_parameters, key):
                setattr(self.wm.server_parameters, key, value)
            else:
                raise AttributeError(
                    "Couldn't find a matching parameter in "
                    "ServerParameters class: '%s'" % key
                )

    def _handle_init(self, msg):
        """
        Deals with initialization messages sent by the server.
        """
        if len(msg) > 2:
            # Player message
            # set the player's uniform number, side, and the play mode as returned
            # by the server directly after connecting.
            side = msg[1]
            uniform_number = msg[2]
            play_mode = msg[3]

            self.wm.side = side
            self.wm.uniform_number = uniform_number
            self.wm.play_mode = play_mode

    def _handle_error(self, msg):
        """
        Deals with error messages by raising them as exceptions.
        """

        m = "Server returned an error: '%s'" % msg[1]
        raise sp_exceptions.SoccerServerError(m)

    def _handle_warning(self, msg):
        """
        Deals with warnings issued by the server.
        """

        m = "Server issued a warning: '%s'" % msg[1]
        print(sp_exceptions.SoccerServerWarning(m))


class ActionHandler:
    """
    Provides facilities for sending commands to the soccer server.  Contains all
    possible commands that can be sent, as well as everything needed to send
    them.  All basic command methods are aliases for placing that command in the
    internal queue and sending it at the appropriate time.
    """

    class CommandType:
        """
        A static class that defines all basic command types.
        """

        # whether the command can only be sent once per cycle or not
        TYPE_PRIMARY = 0
        TYPE_SECONDARY = 1

        # command types corresponding to valid commands to send to the server
        CATCH = "catch"
        CHANGE_VIEW = "change_view"
        DASH = "dash"
        KICK = "kick"
        MOVE = "move"
        SAY = "say"
        SENSE_BODY = "sense_body"
        TURN = "turn"
        TURN_NECK = "turn_neck"

        def __init__(self):
            raise NotImplementedError(
                "Can't instantiate a CommandType, access "
                "its members through ActionHandler instead."
            )

    # a command for our queue containing an id and command text
    Command = collections.namedtuple("Command", "cmd_type text")

    def __init__(self, server_socket):
        """
        Save the socket that connects us to the soccer server to allow us to
        send it commands.
        """

        self.sock = server_socket

        # this contains all requested actions for the current and future cycles
        self.q = queue.Queue()

    def send_commands(self):
        """
        Sends all the enqueued commands.
        """

        # we only send the most recent primary command
        primary_cmd = None

        # dequeue all enqueued commands and send them
        while 1:
            try:
                cmd = self.q.get_nowait()
            except queue.Empty:
                break

            # save the most recent primary command and send it at the very end
            if cmd.cmd_type == ActionHandler.CommandType.TYPE_PRIMARY:
                primary_cmd = cmd
            # send other commands immediately
            else:
                if PRINT_SENT_COMMANDS:
                    print("sent:", cmd.text, "\n")

                self.sock.send(cmd.text)

            # indicate that we finished processing a command
            self.q.task_done()

        # send the saved primary command, if there was one
        if primary_cmd is not None:
            if PRINT_SENT_COMMANDS:
                print("sent:", primary_cmd.text, "\n")

            self.sock.send(primary_cmd.text)
        self.sock.send("(done)")

    def move(self, x, y):
        """
        Teleport the player to some location on the field.  Only works before
        play begins, ie. pre-game, before starting again at half-time, and
        post-goal.  If an invalid location is specified, player is teleported to
        a random location on their side of the field.
        """

        msg = "(move {:.10f} {:.10f})".format(x, y)

        # create the command object for insertion into the queue
        cmd_type = ActionHandler.CommandType.TYPE_PRIMARY
        cmd = ActionHandler.Command(cmd_type, msg)

        self.q.put(cmd)

    def turn(self, relative_degrees):
        """
        Turns the player's body some number of degrees relative to its current
        angle.
        """

        # disallow unreasonable turning
        assert -180 <= relative_degrees <= 180

        msg = "(turn %.10f)" % relative_degrees

        # create the command object for insertion into the queue
        cmd_type = ActionHandler.CommandType.TYPE_PRIMARY
        cmd = ActionHandler.Command(cmd_type, msg)

        self.q.put(cmd)

    def dash(self, power):
        """
        Accelerate the player in the direction its body currently faces.
        """

        msg = "(dash %.10f)" % power

        # create the command object for insertion into the queue
        cmd_type = ActionHandler.CommandType.TYPE_PRIMARY
        cmd = ActionHandler.Command(cmd_type, msg)

        self.q.put(cmd)

    def kick(self, power, relative_direction):
        """
        Accelerates the ball with the given power in the given direction,
        relative to the current direction of the player's body.
        """
        msg = "(kick {:.10f} {:.10f})".format(power, relative_direction)

        # create the command object for insertion into the queue
        cmd_type = ActionHandler.CommandType.TYPE_PRIMARY
        cmd = ActionHandler.Command(cmd_type, msg)

        self.q.put(cmd)

    def catch(self, relative_direction):
        """
        Attempts to catch the ball and put it in the goalie's hand.  The ball
        remains there until the goalie kicks it away.
        """

        msg = "(catch %.10f)" % relative_direction

        # create the command object for insertion into the queue
        cmd_type = ActionHandler.CommandType.TYPE_PRIMARY
        cmd = ActionHandler.Command(cmd_type, msg)

        self.q.put(cmd)

    def say(self, message):
        """
        Says something to other players on the field.  Messages are restricted
        in length, but that isn't enforced here.
        """

        msg = "(say %s)" % message

        # create the command object for insertion into the queue
        cmd_type = ActionHandler.CommandType.TYPE_SECONDARY
        cmd = ActionHandler.Command(cmd_type, msg)

        self.q.put(cmd)

    def turn_neck(self, relative_direction):
        """
        Rotates the player's neck relative to its previous direction.  Neck
        angle is relative to body angle.
        """

        msg = "(turn_neck %.10f)" % relative_direction

        # create the command object for insertion into the queue
        cmd_type = ActionHandler.CommandType.TYPE_SECONDARY
        cmd = ActionHandler.Command(cmd_type, msg)

        self.q.put(cmd)



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""helper functions for logging"""

import abc
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Optional, Union

from acme.utils import loggers, paths
from acme.utils.loggers import base

from mava.utils.loggers.tf_logger import TFSummaryLogger


class MavaLogger(abc.ABC):
    @abc.abstractmethod
    def __init__(
        self,
        label: str,
        *args: Any,
        **kwargs: Any,
    ):
        """Init function that takes in a label and possibly other variables."""

    @abc.abstractmethod
    def write(self, data: Any) -> None:
        """Function that writes logged data."""


class Logger(MavaLogger):
    def __init__(
        self,
        label: str,
        directory: Union[Path, str],
        to_terminal: bool = True,
        to_csv: bool = False,
        to_tensorboard: bool = False,
        time_delta: float = 1.0,
        print_fn: Callable[[str], None] = print,
        time_stamp: Optional[str] = None,
        external_logger: Optional[base.Logger] = None,
        **external_logger_kwargs: Any,
    ):
        self._label = label

        if not isinstance(directory, Path):
            directory = Path(directory)

        self._directory = directory
        self._time_stamp = time_stamp if time_stamp else str(datetime.now())
        self._logger = self.make_logger(
            to_terminal,
            to_csv,
            to_tensorboard,
            time_delta,
            print_fn,
            external_logger=external_logger,
            **external_logger_kwargs,
        )
        self._logger_info = (
            to_terminal,
            to_csv,
            to_tensorboard,
            time_delta,
            print_fn,
            self._time_stamp,
        )

    def update_label(self, label: str) -> None:
        self._label = f"{self._label}_{label}"

    def make_logger(
        self,
        to_terminal: bool,
        to_csv: bool,
        to_tensorboard: bool,
        time_delta: float,
        print_fn: Callable[[str], None],
        external_logger: Optional[base.Logger],
        **external_logger_kwargs: Any,
    ) -> loggers.Logger:
        """Build a Mava logger.

        Args:
            label: Name to give to the logger.
            directory: base directory for the  logging of the experiment.
            to_terminal: to print the logs in the terminal.
            to_csv: to save the logs in a csv file.
            to_tensorboard: to write the logs tf-events.
            time_delta: minimum elapsed time (in seconds) between logging events.
            print_fn: function to call which acts like print.
            external_logger: optional external logger.
            external_logger_kwargs: optional external logger params.
        Returns:
            A logger (pipe) object that responds to logger.write(some_dict).
        """
        logger = []

        if to_terminal:
            logger += [loggers.TerminalLogger(label=self._label, print_fn=print_fn)]

        if to_csv:
            logger += [
                loggers.CSVLogger(
                    directory_or_file=self._path("csv"), label=self._label
                )
            ]

        if to_tensorboard:
            logger += [
                TFSummaryLogger(logdir=self._path("tensorboard"), label=self._label)
            ]

        if external_logger:
            logger += [
                external_logger(
                    label=self._label,
                    **external_logger_kwargs,
                )
            ]

        if logger:
            logger = loggers.Dispatcher(logger)
            logger = loggers.NoneFilter(logger)
            logger = loggers.TimeFilter(logger, time_delta)
        else:
            logger = loggers.NoOpLogger()

        return logger

    def _path(self, subdir: Optional[str] = None) -> str:
        if subdir:
            path = str(self._directory / self._time_stamp / subdir / self._label)
        else:
            path = str(self._directory / self._time_stamp / self._label)

        # Recursively replace "~"
        return paths.process_path(path)

    def write(self, data: Any) -> None:
        self._logger.write(data)



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Code inspired by PyMARL framework implementation
# https://github.com/oxwhirl/pymarl/blob/master/src/modules/mixers/qmix.py

"""Mixing for multi-agent RL systems"""

from typing import Dict

import sonnet as snt
import tensorflow as tf

from mava.components.tf.architectures.base import BaseArchitecture
from mava.components.tf.networks.hypernetwork import HyperNetwork


class MonotonicMixingNetwork(snt.Module):
    """Multi-agent monotonic mixing architecture.
    This is the component which can be used to add monotonic mixing to an underlying
    agent architecture. It currently supports generalised monotonic mixing using
    hypernetworks (1 or 2 layers) for control of decomposition parameters (QMix)."""

    def __init__(
        self,
        architecture: BaseArchitecture,
        agent_networks: Dict[str, snt.Module],
        n_agents: int,
        qmix_hidden_dim: int = 64,
        num_hypernet_layers: int = 2,
        hypernet_hidden_dim: int = 0,
    ) -> None:
        """Initializes the mixer.
        Args:
            state_shape: The state shape as defined by the environment.
            n_agents: The number of agents (i.e. Q-values) to mix.
            qmix_hidden_dim: Mixing layers hidden dimensions.
            num_hypernet_layers: Number of hypernetwork layers. Currently 1 or 2.
            hypernet_hidden_dim: The number of nodes in the hypernetwork hidden
                layer. Relevant for num_hypernet_layers > 1.
        """
        super(MonotonicMixingNetwork, self).__init__()
        self._architecture = architecture
        self._agent_networks = agent_networks
        self._n_agents = n_agents
        self._qmix_hidden_dim = qmix_hidden_dim
        self._num_hypernet_layers = num_hypernet_layers
        self._hypernet_hidden_dim = hypernet_hidden_dim

        # Create hypernetwork
        self._hypernetworks = HyperNetwork(
            self._agent_networks,
            self._qmix_hidden_dim,
            self._num_hypernet_layers,
            self._hypernet_hidden_dim,
        )

    def __call__(
        self,
        q_values: tf.Tensor,  # [batch_size, n_agents]
        states: tf.Tensor,  # [batch_size, state_dim]
    ) -> tf.Tensor:
        """Monotonic mixing logic."""

        # Create hypernetwork
        self._hyperparams = self._hypernetworks(states)

        # Extract hypernetwork layers
        # TODO: make more general -> this assumes two layer hypernetwork
        w1 = self._hyperparams["w1"]  # [B, n_agents, qmix_hidden_dim]
        b1 = self._hyperparams["b1"]  # [B, 1, qmix_hidden_dim]
        w2 = self._hyperparams["w2"]  # [B, qmix_hidden_dim, 1]
        b2 = self._hyperparams["b2"]  # [B, 1, 1]

        # ELU -> Exp. linear unit
        hidden = tf.nn.elu(tf.matmul(q_values, w1) + b1)  # [B, 1, qmix_hidden_dim]

        # Qtot: [B, 1, 1]
        return tf.matmul(hidden, w2) + b2



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A simple multi-agent-system-environment training loop."""

import operator
import time
from typing import Any, Dict, Optional

import acme
import dm_env
import numpy as np
import tree
from acme.utils import counting, loggers
from dm_env import specs

import mava
from mava.utils.wrapper_utils import (
    SeqTimestepDict,
    broadcast_timestep_to_all_agents,
    convert_seq_timestep_and_actions_to_parallel,
    generate_zeros_from_spec,
)


class SequentialEnvironmentLoop(acme.core.Worker):
    """A Sequential MARL environment loop.
    This takes `Environment` and `Executor` instances and coordinates their
    interaction. Executors are updated if `should_update=True`. This can be used as:
        loop = EnvironmentLoop(environment, executor)
        loop.run(num_episodes)
    A `Counter` instance can optionally be given in order to maintain counts
    between different Mava components. If not given a local Counter will be
    created to maintain counts between calls to the `run` method.
    A `Logger` instance can also be passed in order to control the output of the
    loop. If not given a platform-specific default logger will be used as defined
    by utils.loggers.make_default_logger from acme. A string `label` can be passed
    to easily change the label associated with the default logger; this is ignored
    if a `Logger` instance is given.
    """

    def __init__(
        self,
        environment: dm_env.Environment,
        executor: mava.core.Executor,
        counter: counting.Counter = None,
        logger: loggers.Logger = None,
        should_update: bool = True,
        label: str = "sequential_environment_loop",
    ):
        # Internalize agent and environment.
        self._environment = environment
        self._executor = executor
        self._counter = counter or counting.Counter()
        self._logger = logger or loggers.make_default_logger(label)
        self._should_update = should_update
        self._running_statistics = None

    def _get_action(self, agent_id: str, timestep: dm_env.TimeStep) -> Any:
        return self._executor.select_action(agent_id, timestep.observation)

    def run_episode(self) -> loggers.LoggingData:
        """Run one episode.
        Each episode is a loop which interacts first with the environment to get an
        observation and then give that observation to the agent in order to retrieve
        an action.
        Returns:
            An instance of `loggers.LoggingData`.
        """
        # Reset any counts and start the environment.
        start_time = time.time()
        episode_steps = 0

        timestep = self._environment.reset()
        agent = self._environment.current_agent

        # Broadcast timestep for all agents - to use parallel adder.
        # TODO (Kale-ab) : Make more robust -this could cause issues
        # if agents have different discounts, obs or legal actions.
        parallel_timestep = broadcast_timestep_to_all_agents(
            timestep, self._environment.possible_agents
        )

        # Make the first observation - parallel adder.
        self._executor.observe_first(parallel_timestep)

        n_agents = self._environment.num_agents
        rewards = {
            agent: generate_zeros_from_spec(spec)
            for agent, spec in self._environment.reward_spec().items()
        }

        # For evaluation, this keeps track of the total undiscounted reward
        # for each agent accumulated during the episode.
        multiagent_reward_spec = specs.Array((n_agents,), np.float32)
        episode_return = tree.map_structure(
            generate_zeros_from_spec, multiagent_reward_spec
        )

        # Run an episode.
        while not timestep.last():
            # Keep track of timesteps for all agents in the step.
            timesteps: Dict[str, SeqTimestepDict] = {}
            # Generate an action from the agent's policy and step the environment.
            for agent in self._environment.agent_iter(n_agents):
                action = self._get_action(agent, timestep)
                timestep = self._environment.step(action)
                rewards[agent] = timestep.reward

                timesteps[agent] = {"timestep": timestep, "action": action}

            # Combine actions and timesteps to use parallel adder
            (
                parallel_actions,
                parallel_timestep,
            ) = convert_seq_timestep_and_actions_to_parallel(
                timesteps, self._environment.possible_agents
            )

            # Call observe using parallel data.
            self._executor.observe(parallel_actions, next_timestep=parallel_timestep)

            # Update all actors
            if self._should_update:
                self._executor.update()

            # Book-keeping.
            episode_steps += 1

            # Equivalent to: episode_return += timestep.reward
            # We capture the return value because if timestep.reward is a JAX
            # DeviceArray, episode_return will not be mutated in-place. (In all other
            # cases, the returned episode_return will be the same object as the
            # argument episode_return.)
            episode_return = tree.map_structure(
                operator.iadd, episode_return, np.array(list(rewards.values()))
            )

        # Record counts.
        counts = self._counter.increment(episodes=1, steps=episode_steps)

        # Collect the results and combine with counts.
        steps_per_second = episode_steps / (time.time() - start_time)
        result = {
            "episode_length": episode_steps,
            "mean_episode_return": np.mean(episode_return),
            "steps_per_second": steps_per_second,
        }
        result.update(counts)

        return result

    def _compute_step_statistics(self, rewards: Dict[str, float]) -> None:
        pass

    def _compute_episode_statistics(
        self,
        episode_returns: Dict[str, float],
        episode_steps: int,
        start_time: float,
    ) -> None:
        pass

    def run(
        self, num_episodes: Optional[int] = None, num_steps: Optional[int] = None
    ) -> None:
        """Perform the run loop.
        Run the environment loop either for `num_episodes` episodes or for at
        least `num_steps` steps (the last episode is always run until completion,
        so the total number of steps may be slightly more than `num_steps`).
        At least one of these two arguments has to be None.
        Upon termination of an episode a new episode will be started. If the number
        of episodes and the number of steps are not given then this will interact
        with the environment infinitely.
        Args:
            num_episodes: number of episodes to run the loop for.
            num_steps: minimal number of steps to run the loop for.
        Raises:
            ValueError: If both 'num_episodes' and 'num_steps' are not None.
        """

        if not (num_episodes is None or num_steps is None):
            raise ValueError('Either "num_episodes" or "num_steps" should be None.')

        def should_terminate(episode_count: int, step_count: int) -> bool:
            return (num_episodes is not None and episode_count >= num_episodes) or (
                num_steps is not None and step_count >= num_steps
            )

        episode_count, step_count = 0, 0
        while not should_terminate(episode_count, step_count):
            result = self.run_episode()
            episode_count += 1
            step_count += result["episode_length"]
            # Log the given results.
            self._logger.write(result)


class ParallelEnvironmentLoop(acme.core.Worker):
    """A parallel MARL environment loop.
    This takes `Environment` and `Executor` instances and coordinates their
    interaction. Executors are updated if `should_update=True`. This can be used as:
        loop = EnvironmentLoop(environment, executor)
        loop.run(num_episodes)
    A `Counter` instance can optionally be given in order to maintain counts
    between different Mava components. If not given a local Counter will be
    created to maintain counts between calls to the `run` method.
    A `Logger` instance can also be passed in order to control the output of the
    loop. If not given a platform-specific default logger will be used as defined
    by utils.loggers.make_default_logger from acme. A string `label` can be passed
    to easily change the label associated with the default logger; this is ignored
    if a `Logger` instance is given.
    """

    def __init__(
        self,
        environment: dm_env.Environment,
        executor: mava.core.Executor,
        counter: counting.Counter = None,
        logger: loggers.Logger = None,
        should_update: bool = True,
        label: str = "parallel_environment_loop",
    ):
        # Internalize agent and environment.
        self._environment = environment
        self._executor = executor
        self._counter = counter or counting.Counter()
        self._logger = logger or loggers.make_default_logger(label)
        self._should_update = should_update
        self._running_statistics: Dict[str, float] = {}

    def _get_actions(self, timestep: dm_env.TimeStep) -> Any:
        return self._executor.select_actions(timestep.observation)

    def run_episode(self) -> loggers.LoggingData:
        """Run one episode.
        Each episode is a loop which interacts first with the environment to get a
        dictionary of observations and then give those observations to the executor
        in order to retrieve an action for each agent in the system.
        Returns:
            An instance of `loggers.LoggingData`.
        """

        # Reset any counts and start the environment.
        start_time = time.time()
        episode_steps = 0

        timestep = self._environment.reset()

        if type(timestep) == tuple:
            timestep, env_extras = timestep
        else:
            env_extras = {}

        # Make the first observation.
        self._executor.observe_first(timestep, extras=env_extras)

        # For evaluation, this keeps track of the total undiscounted reward
        # for each agent accumulated during the episode.
        rewards: Dict[str, float] = {}
        episode_returns: Dict[str, float] = {}
        for agent, spec in self._environment.reward_spec().items():
            rewards.update({agent: generate_zeros_from_spec(spec)})
            episode_returns.update({agent: generate_zeros_from_spec(spec)})

        # Run an episode.
        while not timestep.last():

            # Generate an action from the agent's policy and step the environment.
            actions = self._get_actions(timestep)

            if type(actions) == tuple:
                # Return other action information
                # e.g. the policy information.
                env_actions, _ = actions
            else:
                env_actions = actions

            timestep = self._environment.step(env_actions)

            if type(timestep) == tuple:
                timestep, env_extras = timestep
            else:
                env_extras = {}

            rewards = timestep.reward

            # Have the agent observe the timestep and let the actor update itself.
            self._executor.observe(
                actions, next_timestep=timestep, next_extras=env_extras
            )

            if self._should_update:
                self._executor.update()

            # Book-keeping.
            episode_steps += 1

            # If env returns empty dict at end of episode.
            if not rewards:
                rewards = {
                    agent: generate_zeros_from_spec(spec)
                    for agent, spec in self._environment.reward_spec().items()
                }

            self._compute_step_statistics(rewards)

            for agent, reward in rewards.items():
                episode_returns[agent] = episode_returns[agent] + reward

        self._compute_episode_statistics(
            episode_returns,
            episode_steps,
            start_time,
        )
        if self._running_statistics:
            return self._running_statistics
        else:
            # Record counts.
            counts = self._counter.increment(episodes=1, steps=episode_steps)

            # Collect the results and combine with counts.
            steps_per_second = episode_steps / (time.time() - start_time)
            result = {
                "episode_length": episode_steps,
                "mean_episode_return": np.mean(list(episode_returns.values())),
                "steps_per_second": steps_per_second,
            }
            result.update(counts)

            return result

    def _compute_step_statistics(self, rewards: Dict[str, float]) -> None:
        pass

    def _compute_episode_statistics(
        self,
        episode_returns: Dict[str, float],
        episode_steps: int,
        start_time: float,
    ) -> None:
        pass

    def run(
        self, num_episodes: Optional[int] = None, num_steps: Optional[int] = None
    ) -> None:
        """Perform the run loop.
        Run the environment loop either for `num_episodes` episodes or for at
        least `num_steps` steps (the last episode is always run until completion,
        so the total number of steps may be slightly more than `num_steps`).
        At least one of these two arguments has to be None.
        Upon termination of an episode a new episode will be started. If the number
        of episodes and the number of steps are not given then this will interact
        with the environment infinitely.
        Args:
            num_episodes: number of episodes to run the loop for.
            num_steps: minimal number of steps to run the loop for.
        Raises:
            ValueError: If both 'num_episodes' and 'num_steps' are not None.
        """

        if not (num_episodes is None or num_steps is None):
            raise ValueError('Either "num_episodes" or "num_steps" should be None.')

        def should_terminate(episode_count: int, step_count: int) -> bool:
            return (num_episodes is not None and episode_count >= num_episodes) or (
                num_steps is not None and step_count >= num_steps
            )

        episode_count, step_count = 0, 0
        while not should_terminate(episode_count, step_count):
            result = self.run_episode()
            episode_count += 1
            step_count += result["episode_length"]
            # Log the given results.
            self._logger.write(result)


# Internal class 1.
# Internal class 2.



# !/usr/bin/env python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# type: ignore


import threading
import time

import numpy as np

from mava.utils.environments.RoboCup_env.robocup_utils import (
    handler,
    sock,
    sp_exceptions,
)
from mava.utils.environments.RoboCup_env.robocup_utils.player_world_model import (
    WorldModel,
)
from mava.utils.environments.RoboCup_env.robocup_utils.util_functions import (
    SpecWrapper,
    wait_for_next_observations,
)

# Define some colours
green = (0, 255, 50)
light_blue = (0, 255, 255)
blue = (0, 0, 255)
red = (255, 0, 0)
yellow = (255, 255, 0)
white = (255, 255, 255)

# R G B
field_col = (0, 200, 50)
team_player = blue
opp_player = light_blue
own_goal = (255, 140, 0)
opp_goal = (128, 0, 128)


class Agent:
    def __init__(self, teamname, team_id, agent_id, num_players, agent_controller=None):
        # whether we're connected to a server yet or not
        self.__connected = False

        self.teamname = teamname
        self.agent_id = agent_id
        self.team_id = team_id
        self.agent_controller = agent_controller

        if self.agent_controller is not None:
            self.specWrapper = SpecWrapper(num_players=num_players)

        # set all variables and important objects to appropriate values for
        # pre-connect state.

        # the socket used to communicate with the server
        self.__sock = None

        # models and the message handler for parsing and storing information
        self.wm = None
        self.msg_handler = None

        # parse thread and control variable
        self.__parsing = False
        self.__msg_thread = None

        self.__thinking = False  # think thread and control variable
        self.__think_thread = None

        # whether we should send commands
        self.__send_commands = False

        # adding goal post markers
        self.enemy_goal_pos = None
        self.own_goal_pos = None

    def connect(self, host, port, version=11):
        """
        Gives us a connection to the server as one player on a team.  This
        immediately connects the agent to the server and starts receiving and
        parsing the information it sends.
        """
        self.start_comm_threads(host, port, self.teamname, version)

    def start_comm_threads(self, host, port, teamname, version):
        # if already connected, raise an error since user may have wanted to
        # connect again to a different server.
        if self.__connected:
            msg = "Cannot connect while already connected, disconnect first."
            raise sp_exceptions.AgentConnectionStateError(msg)

        # the pipe through which all of our communication takes place
        # p rint("Start comms: ", port)
        self.__sock = sock.Socket(host, port)

        # our models of the world and our body
        self.wm = WorldModel(handler.ActionHandler(self.__sock))

        # set the team name of the world model to the given name
        self.wm.teamname = teamname

        # handles all messages received from the server
        self.msg_handler = handler.MessageHandler(self.wm)

        # set up our threaded message receiving system
        self.__parsing = True  # tell thread that we're currently running

        init_address = self.__sock.address
        # send the init message and allow the message handler to handle further
        # responses.
        init_msg = "(init %s (version %d))"
        self.__sock.send(init_msg % (teamname, version))

        self.__message_tread = threading.Thread(
            target=self.__message_loop, name="message_loop"
        )
        self.__message_tread.daemon = True
        self.__message_tread.start()
        # wait until the socket receives a response from the server and gets its
        # assigned port.
        while self.__sock.address == init_address:
            time.sleep(0.0001)

        # create our thinking thread.  this will perform the actions necessary
        # to play a game of robo-soccer.
        self.__thinking = True

        # set connected state.  done last to prevent state inconsistency if
        # something goes wrong beforehand.
        self.__connected = True

        self.last_action = None

        # Move to random starting position
        x = np.random.randint(-52, 53)
        y = np.random.randint(-34, 35)

        self.__sock.send("(move " + str(x) + " " + str(y) + ")")
        self.__sock.send("(done)")

        if self.agent_controller is not None:
            self.__think_thread = threading.Thread(
                target=self.__think_loop, name="think_loop"
            )
            self.__think_thread.daemon = True
            self.__think_thread.start()

    def __message_loop(self):
        """
        Handles messages received from the server.

        This SHOULD NOT be called externally, since it's used as a threaded loop
        internally by this object.  Calling it externally is a BAD THING!
        """

        # loop until we're told to stop
        while self.__parsing:
            # receive message data from the server and pass it along to the
            # world model as-is.  the world model parses it and stores it within
            # itself for perusal at our leisure.
            raw_msg = self.__sock.recv()

            msg_type = self.msg_handler.handle_message(raw_msg)
            if msg_type is not None:
                # we send commands all at once every cycle, ie. whenever a
                # 'sense_body' command is received
                if msg_type == handler.ActionHandler.CommandType.SENSE_BODY:
                    self.__send_commands = True

    def __think_loop(self):
        """
        Performs world model analysis and sends appropriate commands to the
        server to allow the agent to participate in the current game.

        Like the message loop, this SHOULD NOT be called externally.  Use the
        play method to start play, and the disconnect method to end it.
        """

        while self.__thinking:
            # Get and process the latest observation
            observation = self.specWrapper.proc_agent_env_obs(
                env_agent_obs=self.get_latest_observation(),
                last_action=self.last_action,
            )

            # Get the agent's action and covert it to a RoboCup environment command.

            self.last_action = self.agent_controller.get_action(observation)
            action = self.specWrapper.proc_agent_action(self.last_action)

            # Send the action
            self.__sock.send(action)
            self.__sock.send("(done)")

            # Wait for a new timestep indicator
            wait_for_next_observations([self])

    def get_latest_observation(self):
        # Generate new observation
        obs_dict = {}
        obs_dict["players"] = self.wm.players

        # Ball
        obs_dict["ball"] = self.wm.ball

        # Team side
        obs_dict["side"] = self.wm.side

        # Score
        obs_dict["scores"] = [self.wm.score_l, self.wm.score_r]

        obs_dict["view_width"] = self.wm.view_width
        obs_dict["view_quality"] = self.wm.view_quality
        obs_dict["stamina"] = self.wm.stamina
        obs_dict["effort"] = self.wm.effort
        obs_dict["speed_amount"] = self.wm.speed_amount

        # TODO: Is these two measurements relative? I think they are.
        obs_dict["speed_direction"] = self.wm.speed_direction
        obs_dict["neck_direction"] = self.wm.neck_direction

        # self.wm.obs_updated = False
        obs_dict["estimated_abs_coords"] = self.wm.abs_coords
        obs_dict["estimated_abs_body_dir"] = self.wm.abs_body_dir
        obs_dict["estimated_abs_neck_dir"] = self.wm.abs_neck_dir

        # Fix observations so that it is side almost agnostic. A agent can
        # be trained on one side and still execute what it has learned on
        # the other side of the field.
        if obs_dict["side"] == "r":
            if obs_dict["estimated_abs_coords"][0] is not None:
                obs_dict["estimated_abs_coords"] = [
                    -obs_dict["estimated_abs_coords"][0],
                    -obs_dict["estimated_abs_coords"][1],
                ]
                obs_dict["estimated_abs_neck_dir"] += 180
                obs_dict["estimated_abs_body_dir"] += 180

        else:
            assert obs_dict["side"] == "l"

        return obs_dict

    def do_action(self, action):
        # tell the ActionHandler to send its enqueued messages if it is time
        # only think if new data has arrived
        if self.agent_controller:
            action = self.agent_controller.get_action(
                self.get_latest_observation(), None
            )
            print("When is this actually used?")
            exit()

        self.__sock.send(action)
        self.__sock.send("(done)")



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""VDN trainer implementation."""

from typing import Any, Dict, List, Optional, Union

import sonnet as snt
import tensorflow as tf
from acme.utils import counting, loggers
from trfl.indexing_ops import batched_index

from mava.components.tf.modules.communication import BaseCommunicationModule
from mava.components.tf.modules.exploration.exploration_scheduling import (
    LinearExplorationScheduler,
)
from mava.systems.tf.madqn.training import MADQNTrainer
from mava.utils import training_utils as train_utils

train_utils.set_growing_gpu_memory()


class VDNTrainer(MADQNTrainer):
    """VDN trainer.
    This is the trainer component of a VDN system. i.e. it takes a dataset as input
    and implements update functionality to learn from this dataset.
    """

    def __init__(
        self,
        agents: List[str],
        agent_types: List[str],
        q_networks: Dict[str, snt.Module],
        target_q_networks: Dict[str, snt.Module],
        mixing_network: snt.Module,
        target_mixing_network: snt.Module,
        target_update_period: int,
        dataset: tf.data.Dataset,
        optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]],
        discount: float,
        shared_weights: bool,
        exploration_scheduler: LinearExplorationScheduler,
        communication_module: Optional[BaseCommunicationModule] = None,
        max_gradient_norm: float = None,
        counter: counting.Counter = None,
        fingerprint: bool = False,
        logger: loggers.Logger = None,
        checkpoint: bool = True,
        checkpoint_subpath: str = "~/mava/",
    ) -> None:

        self._mixing_network = mixing_network
        self._target_mixing_network = target_mixing_network

        super(VDNTrainer, self).__init__(
            agents=agents,
            agent_types=agent_types,
            q_networks=q_networks,
            target_q_networks=target_q_networks,
            target_update_period=target_update_period,
            dataset=dataset,
            optimizer=optimizer,
            discount=discount,
            shared_weights=shared_weights,
            exploration_scheduler=exploration_scheduler,
            communication_module=communication_module,
            max_gradient_norm=max_gradient_norm,
            counter=counter,
            fingerprint=fingerprint,
            logger=logger,
            checkpoint=checkpoint,
            checkpoint_subpath=checkpoint_subpath,
        )

    @tf.function
    def _step(
        self,
    ) -> Dict[str, Dict[str, Any]]:

        # Update the target networks
        self._update_target_networks()

        # Get data from replay (dropping extras if any). Note there is no
        # extra data here because we do not insert any into Reverb.
        inputs = next(self._iterator)

        self._forward(inputs)

        self._backward()

        # Log losses per agent
        return {agent: {"q_value_loss": self.loss} for agent in self._agents}

    def _forward(self, inputs: Any) -> None:
        # Unpack input data as follows:
        # o_tm1 = dictionary of observations one for each agent
        # a_tm1 = dictionary of actions taken from obs in o_tm1
        # e_tm1 [Optional] = extra data that the agents persist in replay.
        # r_t = dictionary of rewards or rewards sequences
        #   (if using N step transitions) ensuing from actions a_tm1
        # d_t = environment discount ensuing from actions a_tm1.
        #   This discount is applied to future rewards after r_t.
        # o_t = dictionary of next observations or next observation sequences
        # e_t = [Optional] = extra data that the agents persist in replay.
        o_tm1, a_tm1, _, r_t, d_t, o_t, _ = inputs.data

        # Do forward passes through the networks and calculate the losses
        with tf.GradientTape(persistent=True) as tape:
            q_acts = []  # Q vals
            q_targets = []  # Target Q vals
            for agent in self._agents:
                agent_key = self.agent_net_keys[agent]

                o_tm1_feed, o_t_feed, a_tm1_feed = self._get_feed(
                    o_tm1, o_t, a_tm1, agent
                )
                q_tm1 = self._q_networks[agent_key](o_tm1_feed)
                q_t_value = self._target_q_networks[agent_key](o_t_feed)
                q_t_selector = self._q_networks[agent_key](o_t_feed)
                best_action = tf.argmax(q_t_selector, axis=1, output_type=tf.int32)

                # TODO Make use of q_t_selector for fingerprinting. Speak to Claude.
                q_act = batched_index(q_tm1, a_tm1_feed, keepdims=True)  # [B, 1]
                q_target = batched_index(
                    q_t_value, best_action, keepdims=True
                )  # [B, 1]

                q_acts.append(q_act)
                q_targets.append(q_target)

            rewards = tf.concat(
                [tf.reshape(val, (-1, 1)) for val in list(r_t.values())], axis=1
            )
            rewards = tf.reduce_mean(rewards, axis=1)  # [B]

            pcont = tf.concat(
                [tf.reshape(val, (-1, 1)) for val in list(d_t.values())], axis=1
            )
            pcont = tf.reduce_mean(pcont, axis=1)
            discount = tf.cast(self._discount, list(d_t.values())[0].dtype)
            pcont = discount * pcont  # [B]

            q_acts = tf.concat(q_acts, axis=1)  # [B, num_agents]
            q_targets = tf.concat(q_targets, axis=1)  # [B, num_agents]

            q_tot_mixed = self._mixing_network(q_acts)  # [B, 1, 1]
            q_tot_target_mixed = self._target_mixing_network(q_targets)  # [B, 1, 1]

            # q_tot_mixed = tf.reduce_sum(q_acts, axis=1)  # [B, 1, 1]
            # q_tot_target_mixed = tf.reduce_sum(q_targets, axis=1)  # [B, 1, 1]

            # Calculate Q loss.
            targets = rewards + pcont * q_tot_target_mixed
            targets = tf.stop_gradient(targets)
            td_error = targets - q_tot_mixed

            # Loss is MSE scaled by 0.5, so the gradient is equal to the TD error.
            self.loss = 0.5 * tf.reduce_mean(tf.square(td_error))
            self.tape = tape

    def _backward(self) -> None:
        # Calculate the gradients and update the networks
        for agent in self._agents:
            agent_key = self.agent_net_keys[agent]
            # Get trainable variables.
            trainable_variables = self._q_networks[agent_key].trainable_variables

            # Compute gradients.
            gradients = self.tape.gradient(self.loss, trainable_variables)

            # Clip gradients.
            gradients = tf.clip_by_global_norm(gradients, self._max_gradient_norm)[0]

            # Apply gradients.
            self._optimizers[agent_key].apply(gradients, trainable_variables)

        # Delete the tape manually because of the persistent=True flag.
        train_utils.safe_del(self, "tape")



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DIAL executor implementation."""
from typing import Any, Dict, Optional

import sonnet as snt
import tensorflow as tf
from acme import types
from acme.tf import variable_utils as tf2_variable_utils

from mava import adders
from mava.components.tf.modules.communication import BaseCommunicationModule
from mava.systems.tf.madqn.execution import MADQNRecurrentCommExecutor
from mava.systems.tf.madqn.training import MADQNTrainer


class DIALSwitchExecutor(MADQNRecurrentCommExecutor):
    """DIAL executor.
    An executor based on a recurrent communicating policy for each agent in the system
    which takes non-batched observations and outputs non-batched actions.
    It also allows adding experiences to replay and updating the weights
    from the policy on the learner.
    Note: this executor is specific to switch game env.
    """

    def __init__(
        self,
        q_networks: Dict[str, snt.Module],
        action_selectors: Dict[str, snt.Module],
        communication_module: BaseCommunicationModule,
        shared_weights: bool = True,
        adder: Optional[adders.ParallelAdder] = None,
        variable_client: Optional[tf2_variable_utils.VariableClient] = None,
        store_recurrent_state: bool = True,
        trainer: MADQNTrainer = None,
        fingerprint: bool = False,
        evaluator: bool = False,
    ):
        """Initializes the executor.
        Args:
          policy_network: the policy to run for each agent in the system.
          shared_weights: specify if weights are shared between agent networks.
          adder: the adder object to which allows to add experiences to a
            dataset/replay buffer.
          variable_client: object which allows to copy weights from the trainer copy
            of the policies to the executor copy (in case they are separate).
        """

        # Store these for later use.
        self._adder = adder
        self._variable_client = variable_client
        self._q_networks = q_networks
        self._policy_networks = q_networks
        self._communication_module = communication_module
        self._action_selectors = action_selectors
        self._store_recurrent_state = store_recurrent_state
        self._trainer = trainer
        self._shared_weights = shared_weights

        self._states: Dict[str, Any] = {}
        self._messages: Dict[str, Any] = {}

    def _policy(
        self,
        agent: str,
        observation: types.NestedTensor,
        state: types.NestedTensor,
        message: types.NestedTensor,
        legal_actions: types.NestedTensor,
        epsilon: tf.Tensor,
    ) -> types.NestedTensor:

        (action, m_values), new_state = super()._policy(
            agent,
            observation,
            state,
            message,
            legal_actions,
            epsilon,
        )

        # Mask message if obs[0] == 1.
        # Note: this is specific to switch env
        if observation[0] == 0:
            m_values = tf.zeros_like(m_values)

        return (action, m_values), new_state



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# type: ignore


class SoccerServerError(Exception):
    """
    Represents an error message returned by the soccer server.
    """


class SoccerServerWarning(Exception):
    """
    Represents a warning message returned by the soccer server.
    """


class MessageTypeError(Exception):
    """
    An exception for an unknown message type received from the soccer server.
    """


class AgentAlreadyPlayingError(Exception):
    """
    Raised when a user calls an agent's play method after it has already started
    playing.
    """


class ObjectTypeError(Exception):
    """
    Raised when an unknown object type is encountered in a sense message.
    """


class AgentConnectionStateError(Exception):
    """
    Raised when methods are called at an inappropriate time relative to the
    connection state of the agent object.
    """



"""mava is a framework for multi-agent reinforcement learning."""

# Expose specs and types modules.
from acme import types as acme_types

# Internal core import.
from acme.core import Saveable, VariableSource, Worker

from mava import specs, types, utils

# Make __version__ accessible.
from mava._metadata import __version__

# Expose core interfaces.
from mava.core import Executor, Trainer

# Expose the environment loop.
from mava.environment_loop import ParallelEnvironmentLoop, SequentialEnvironmentLoop
from mava.specs import MAEnvironmentSpec

# Internal environment_loop import.

# Mava loves you more. ;)



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""DIAL trainer implementation."""

from typing import Any, Dict, List, Union

import sonnet as snt
import tensorflow as tf
import tree
import trfl
from acme.tf import utils as tf2_utils
from acme.types import NestedArray
from acme.utils import counting, loggers

from mava.components.tf.modules.communication import BaseCommunicationModule
from mava.components.tf.modules.exploration.exploration_scheduling import (
    LinearExplorationScheduler,
)
from mava.systems.tf.madqn.training import MADQNRecurrentCommTrainer
from mava.utils import training_utils as train_utils

train_utils.set_growing_gpu_memory()


class DIALSwitchTrainer(MADQNRecurrentCommTrainer):
    """Recurrent Comm DIAL Switch trainer.
    This is the trainer component of a DIAL system. IE it takes a dataset as input
    and implements update functionality to learn from this dataset.
    Note: this trainer is specific to switch game env.
    """

    def __init__(
        self,
        agents: List[str],
        agent_types: List[str],
        q_networks: Dict[str, snt.Module],
        target_q_networks: Dict[str, snt.Module],
        target_update_period: int,
        dataset: tf.data.Dataset,
        optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]],
        discount: float,
        shared_weights: bool,
        exploration_scheduler: LinearExplorationScheduler,
        communication_module: BaseCommunicationModule,
        max_gradient_norm: float = None,
        fingerprint: bool = False,
        counter: counting.Counter = None,
        logger: loggers.Logger = None,
        checkpoint: bool = True,
        checkpoint_subpath: str = "~/mava/",
    ):
        super().__init__(
            agents=agents,
            agent_types=agent_types,
            q_networks=q_networks,
            target_q_networks=target_q_networks,
            target_update_period=target_update_period,
            dataset=dataset,
            optimizer=optimizer,
            discount=discount,
            shared_weights=shared_weights,
            exploration_scheduler=exploration_scheduler,
            max_gradient_norm=max_gradient_norm,
            fingerprint=fingerprint,
            counter=counter,
            logger=logger,
            checkpoint=checkpoint,
            checkpoint_subpath=checkpoint_subpath,
            communication_module=communication_module,
        )

    def _forward(self, inputs: Any) -> None:
        data = tree.map_structure(
            lambda v: tf.expand_dims(v, axis=0) if len(v.shape) <= 1 else v, inputs.data
        )
        data = tf2_utils.batch_to_sequence(data)

        observations, actions, rewards, discounts, _, extra = data

        # core_states = extra["core_states"]
        core_state = tree.map_structure(
            lambda s: s[:, 0, :], inputs.data.extras["core_states"]
        )
        core_message = tree.map_structure(
            lambda s: s[:, 0, :], inputs.data.extras["core_messages"]
        )
        T = actions[self._agents[0]].shape[0]

        # Use fact that end of episode always has the reward to
        # find episode lengths. This is used to mask loss.
        ep_end = tf.argmax(tf.math.abs(rewards[self._agents[0]]), axis=0)

        with tf.GradientTape(persistent=True) as tape:
            q_network_losses: Dict[str, NestedArray] = {
                agent: {"q_value_loss": tf.zeros(())} for agent in self._agents
            }

            state = {agent: core_state[agent][0] for agent in self._agents}
            target_state = {agent: core_state[agent][0] for agent in self._agents}

            message = {agent: core_message[agent][0] for agent in self._agents}
            target_message = {agent: core_message[agent][0] for agent in self._agents}

            # _target_q_networks must be 1 step ahead
            target_channel = self._communication_module.process_messages(target_message)
            for agent in self._agents:
                agent_key = self.agent_net_keys[agent]
                (q_targ, m), s = self._target_q_networks[agent_key](
                    observations[agent].observation[0],
                    target_state[agent],
                    target_channel[agent],
                )
                target_state[agent] = s
                target_message[agent] = m

            for t in range(1, T, 1):
                channel = self._communication_module.process_messages(message)
                target_channel = self._communication_module.process_messages(
                    target_message
                )

                for agent in self._agents:
                    agent_key = self.agent_net_keys[agent]

                    # Cast the additional discount
                    # to match the environment discount dtype.

                    discount = tf.cast(self._discount, dtype=discounts[agent][0].dtype)

                    (q_targ, m), s = self._target_q_networks[agent_key](
                        observations[agent].observation[t],
                        target_state[agent],
                        target_channel[agent],
                    )

                    target_state[agent] = s
                    target_message[agent] = tf.math.multiply(
                        m, observations[agent].observation[t][:, :1]
                    )

                    (q, m), s = self._q_networks[agent_key](
                        observations[agent].observation[t - 1],
                        state[agent],
                        channel[agent],
                    )

                    state[agent] = s
                    message[agent] = tf.math.multiply(
                        m, observations[agent].observation[t - 1][:, :1]
                    )

                    # Mask target
                    q_targ = tf.concat(
                        [
                            [q_targ[i]]
                            if t <= ep_end[i]
                            else [tf.zeros_like(q_targ[i])]
                            for i in range(q_targ.shape[0])
                        ],
                        axis=0,
                    )

                    loss, _ = trfl.qlearning(
                        q,
                        actions[agent][t - 1],
                        rewards[agent][t - 1],
                        discount * discounts[agent][t],
                        q_targ,
                    )

                    # Index loss (mask ended episodes)
                    if not tf.reduce_any(t - 1 <= ep_end):
                        continue

                    loss = tf.reduce_mean(loss[t - 1 <= ep_end])
                    # loss = tf.reduce_mean(loss)
                    q_network_losses[agent]["q_value_loss"] += loss

        # print(q_network_losses['agent_0'])
        self._q_network_losses = q_network_losses
        self.tape = tape
        # {}.t



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines the DIAL system class."""
import functools
from typing import Any, Callable, Dict, Optional, Type, Union

import acme
import dm_env
import launchpad as lp
import reverb
import sonnet as snt
from acme import specs as acme_specs
from acme.tf import utils as tf2_utils
from acme.utils import counting

import mava
from mava import core
from mava import specs as mava_specs
from mava.components.tf.architectures import DecentralisedValueActor
from mava.components.tf.modules.communication import (
    BaseCommunicationModule,
    BroadcastedCommunication,
)
from mava.components.tf.modules.exploration import LinearExplorationScheduler
from mava.components.tf.modules.stabilising import FingerPrintStabalisation
from mava.environment_loop import ParallelEnvironmentLoop
from mava.systems.tf import executors
from mava.systems.tf import savers as tf2_savers
from mava.systems.tf.dial import builder
from mava.systems.tf.madqn import execution, training
from mava.utils import lp_utils
from mava.utils.loggers import MavaLogger, logger_utils
from mava.wrappers import DetailedPerAgentStatistics


class DIAL:
    """DIAL system.
    This implements a single-process QMIX system.
    Args:
        environment_factory: Callable to instantiate an environment on a compute node.
        network_factory: Callable to instantiate system networks on a compute node.
        logger_factory: Callable to instantiate a system logger on a compute node.
        architecture: system architecture, e.g. decentralised or centralised.
        trainer_fn: training type associated with executor and architecture,
            e.g. centralised training.
        executor_fn: executor type for example feedforward or recurrent.
        num_executors: number of executor processes to run in parallel.
        num_caches: number of trainer node caches.
        environment_spec: description of the actions, observations, etc.
        q_networks: the online Q network (the one being optimized)
        epsilon: probability of taking a random action; ignored if a policy
            network is given.
        trainer_fn: the class used for training the agent and mixing networks.
        shared_weights: boolean determining whether shared weights is used.
        target_update_period: number of learner steps to perform before updating
            the target networks.
        clipping: whether to clip gradients by global norm.
        replay_table_name: string indicating what name to give the replay table.
        max_replay_size: maximum replay size.
        samples_per_insert: number of samples to take from replay for every insert
            that is made.
        prefetch_size: size to prefetch from replay.
        batch_size: batch size for updates.
        n_step: number of steps to squash into a single transition.
        discount: discount to use for TD updates.
        counter: counter object used to keep track of steps.
        checkpoint: boolean indicating whether to checkpoint the learner.
    """

    def __init__(
        self,
        environment_factory: Callable[[bool], dm_env.Environment],
        network_factory: Callable[[acme_specs.BoundedArray], Dict[str, snt.Module]],
        logger_factory: Callable[[str], MavaLogger] = None,
        architecture: Type[DecentralisedValueActor] = DecentralisedValueActor,
        trainer_fn: Type[
            training.MADQNRecurrentCommTrainer
        ] = training.MADQNRecurrentCommTrainer,
        communication_module: Type[BaseCommunicationModule] = BroadcastedCommunication,
        executor_fn: Type[core.Executor] = execution.MADQNFeedForwardExecutor,
        exploration_scheduler_fn: Type[
            LinearExplorationScheduler
        ] = LinearExplorationScheduler,
        replay_stabilisation_fn: Optional[Type[FingerPrintStabalisation]] = None,
        epsilon_min: float = 0.05,
        epsilon_decay: float = 1e-3,
        num_executors: int = 1,
        num_caches: int = 0,
        environment_spec: mava_specs.MAEnvironmentSpec = None,
        shared_weights: bool = True,
        batch_size: int = 256,
        prefetch_size: int = 4,
        min_replay_size: int = 1000,
        max_replay_size: int = 1000000,
        samples_per_insert: Optional[float] = 4.0,
        n_step: int = 5,
        sequence_length: int = 6,
        period: int = 20,
        max_gradient_norm: float = None,
        discount: float = 1,
        optimizer: Union[snt.Optimizer, Dict[str, snt.Optimizer]] = snt.optimizers.Adam(
            learning_rate=1e-4
        ),
        target_update_period: int = 100,
        executor_variable_update_period: int = 1000,
        max_executor_steps: int = None,
        checkpoint: bool = False,
        checkpoint_subpath: str = "~/mava/",
        logger_config: Dict = {},
        train_loop_fn: Callable = ParallelEnvironmentLoop,
        eval_loop_fn: Callable = ParallelEnvironmentLoop,
        train_loop_fn_kwargs: Dict = {},
        eval_loop_fn_kwargs: Dict = {},
    ):

        if not environment_spec:
            environment_spec = mava_specs.MAEnvironmentSpec(
                environment_factory(evaluation=False)  # type:ignore
            )

        # set default logger if no logger provided
        if not logger_factory:
            logger_factory = functools.partial(
                logger_utils.make_logger,
                directory="~/mava",
                to_terminal=True,
                time_delta=10,
            )

        self._architecture = architecture
        self._communication_module_fn = communication_module
        self._environment_factory = environment_factory
        self._network_factory = network_factory
        self._logger_factory = logger_factory
        self._environment_spec = environment_spec
        self._shared_weights = shared_weights
        self._num_exectors = num_executors
        self._num_caches = num_caches
        self._max_executor_steps = max_executor_steps
        self._checkpoint_subpath = checkpoint_subpath
        self._checkpoint = checkpoint
        self._logger_config = logger_config
        self._train_loop_fn = train_loop_fn
        self._train_loop_fn_kwargs = train_loop_fn_kwargs
        self._eval_loop_fn = eval_loop_fn
        self._eval_loop_fn_kwargs = eval_loop_fn_kwargs

        if issubclass(executor_fn, executors.RecurrentExecutor):
            extra_specs = self._get_extra_specs()
        else:
            extra_specs = {}

        self._builder = builder.DIALBuilder(
            builder.DIALConfig(
                environment_spec=environment_spec,
                epsilon_min=epsilon_min,
                epsilon_decay=epsilon_decay,
                shared_weights=shared_weights,
                discount=discount,
                batch_size=batch_size,
                prefetch_size=prefetch_size,
                target_update_period=target_update_period,
                executor_variable_update_period=executor_variable_update_period,
                min_replay_size=min_replay_size,
                max_replay_size=max_replay_size,
                samples_per_insert=samples_per_insert,
                n_step=n_step,
                sequence_length=sequence_length,
                period=period,
                max_gradient_norm=max_gradient_norm,
                checkpoint=checkpoint,
                optimizer=optimizer,
                checkpoint_subpath=checkpoint_subpath,
            ),
            trainer_fn=trainer_fn,
            executor_fn=executor_fn,
            extra_specs=extra_specs,
            exploration_scheduler_fn=exploration_scheduler_fn,
            replay_stabilisation_fn=replay_stabilisation_fn,
        )

    def _get_extra_specs(self) -> Any:
        agents = self._environment_spec.get_agent_ids()
        core_state_specs = {}
        core_message_specs = {}

        networks = self._network_factory(  # type: ignore
            environment_spec=self._environment_spec
        )
        for agent in agents:
            agent_type = agent.split("_")[0]
            core_state_specs[agent] = (
                tf2_utils.squeeze_batch_dim(
                    networks["q_networks"][agent_type].initial_state(1)
                ),
            )
            if self._communication_module_fn is not None:
                core_message_specs[agent] = (
                    tf2_utils.squeeze_batch_dim(
                        networks["q_networks"][agent_type].initial_message(1)
                    ),
                )

        extras = {
            "core_states": core_state_specs,
            "core_messages": core_message_specs,
        }
        return extras

    def replay(self) -> Any:
        """The replay storage."""
        return self._builder.make_replay_tables(self._environment_spec)

    def counter(self) -> Any:
        return tf2_savers.CheckpointingRunner(
            counting.Counter(),
            time_delta_minutes=15,
            directory=self._checkpoint_subpath,
            subdirectory="counter",
        )

    def coordinator(self, counter: counting.Counter) -> Any:
        return lp_utils.StepsLimiter(counter, self._max_executor_steps)  # type: ignore

    def trainer(
        self,
        replay: reverb.Client,
        counter: counting.Counter,
    ) -> mava.core.Trainer:
        """The Trainer part of the system."""

        # Create the networks to optimize (online)
        networks = self._network_factory(  # type: ignore
            environment_spec=self._environment_spec, shared_weights=self._shared_weights
        )

        # Create system architecture with target networks.
        architecture = self._architecture(
            environment_spec=self._environment_spec,
            value_networks=networks["q_networks"],
            shared_weights=self._shared_weights,
        )

        if self._builder._replay_stabiliser_fn is not None:
            architecture = self._builder._replay_stabiliser_fn(  # type: ignore
                architecture
            )

        communication_module = None
        if self._communication_module_fn is not None:
            communication_module = self._communication_module_fn(
                architecture=architecture,
                shared=True,
                channel_size=1,
                channel_noise=0,
            )
            system_networks = communication_module.create_system()
        else:
            system_networks = architecture.create_system()

        # create logger
        trainer_logger_config = {}
        if self._logger_config and "trainer" in self._logger_config:
            trainer_logger_config = self._logger_config["trainer"]
        trainer_logger = self._logger_factory(  # type: ignore
            "trainer", **trainer_logger_config
        )

        dataset = self._builder.make_dataset_iterator(replay)
        counter = counting.Counter(counter, "trainer")

        return self._builder.make_trainer(
            networks=system_networks,
            dataset=dataset,
            counter=counter,
            communication_module=communication_module,
            logger=trainer_logger,
        )

    def executor(
        self,
        executor_id: str,
        replay: reverb.Client,
        variable_source: acme.VariableSource,
        counter: counting.Counter,
        trainer: Optional[training.MADQNRecurrentCommTrainer] = None,
    ) -> mava.ParallelEnvironmentLoop:
        """The executor process."""

        # Create the behavior policy.
        networks = self._network_factory(  # type: ignore
            environment_spec=self._environment_spec, shared_weights=self._shared_weights
        )

        # Create system architecture with target networks.
        architecture = self._architecture(
            environment_spec=self._environment_spec,
            value_networks=networks["q_networks"],
            shared_weights=self._shared_weights,
        )

        if self._builder._replay_stabiliser_fn is not None:
            architecture = self._builder._replay_stabiliser_fn(  # type: ignore
                architecture
            )

        communication_module = None
        if self._communication_module_fn is not None:
            communication_module = self._communication_module_fn(
                architecture=architecture,
                shared=True,
                channel_size=1,
                channel_noise=0,
            )
            system_networks = communication_module.create_system()
        else:
            system_networks = architecture.create_system()

        # Create the executor.
        executor = self._builder.make_executor(
            q_networks=system_networks["values"],
            action_selectors=networks["action_selectors"],
            communication_module=communication_module,
            adder=self._builder.make_adder(replay),
            variable_source=variable_source,
            trainer=trainer,
        )

        # TODO (Arnu): figure out why factory function are giving type errors
        # Create the environment.
        environment = self._environment_factory(evaluation=False)  # type: ignore

        # Create logger and counter; actors will not spam bigtable.
        counter = counting.Counter(counter, "executor")

        # Create executor logger
        executor_logger_config = {}
        if self._logger_config and "executor" in self._logger_config:
            executor_logger_config = self._logger_config["executor"]
        exec_logger = self._logger_factory(  # type: ignore
            f"executor_{executor_id}", **executor_logger_config
        )

        # Create the loop to connect environment and executor.
        train_loop = self._train_loop_fn(
            environment,
            executor,
            counter=counter,
            logger=exec_logger,
            **self._train_loop_fn_kwargs,
        )

        train_loop = DetailedPerAgentStatistics(train_loop)

        return train_loop

    def evaluator(
        self,
        variable_source: acme.VariableSource,
        counter: counting.Counter,
        trainer: training.MADQNRecurrentCommTrainer,
    ) -> Any:
        """The evaluation process."""

        # Create the behavior policy.
        networks = self._network_factory(  # type: ignore
            environment_spec=self._environment_spec, shared_weights=self._shared_weights
        )

        # Create system architecture with target networks.
        architecture = self._architecture(
            environment_spec=self._environment_spec,
            value_networks=networks["q_networks"],
            shared_weights=self._shared_weights,
        )

        if self._builder._replay_stabiliser_fn is not None:
            architecture = self._builder._replay_stabiliser_fn(  # type: ignore
                architecture
            )

        communication_module = None
        if self._communication_module_fn is not None:
            communication_module = self._communication_module_fn(
                architecture=architecture,
                shared=True,
                channel_size=1,
                channel_noise=0,
            )
            system_networks = communication_module.create_system()
        else:
            system_networks = architecture.create_system()

        # Create the agent.
        executor = self._builder.make_executor(
            q_networks=system_networks["values"],
            action_selectors=networks["action_selectors"],
            variable_source=variable_source,
            communication_module=communication_module,
            trainer=trainer,
            evaluator=True,
        )

        # Make the environment.
        environment = self._environment_factory(evaluation=True)  # type: ignore

        # Create logger and counter.
        counter = counting.Counter(counter, "evaluator")
        evaluator_logger_config = {}
        if self._logger_config and "evaluator" in self._logger_config:
            evaluator_logger_config = self._logger_config["evaluator"]
        eval_logger = self._logger_factory(  # type: ignore
            "evaluator", **evaluator_logger_config
        )

        # Create the run loop and return it.
        # Create the loop to connect environment and executor.
        eval_loop = self._eval_loop_fn(
            environment,
            executor,
            counter=counter,
            logger=eval_logger,
            **self._eval_loop_fn_kwargs,
        )

        eval_loop = DetailedPerAgentStatistics(eval_loop)
        return eval_loop

    def build(self, name: str = "madqn") -> Any:
        """Build the distributed system topology."""
        program = lp.Program(name=name)
        counter = None

        with program.group("replay"):
            replay = program.add_node(lp.ReverbNode(self.replay))

        if self._checkpoint:
            with program.group("counter"):
                counter = program.add_node(lp.CourierNode(self.counter))

        if self._max_executor_steps:
            with program.group("coordinator"):
                _ = program.add_node(lp.CourierNode(self.coordinator, counter))

        with program.group("trainer"):
            trainer = program.add_node(lp.CourierNode(self.trainer, replay, counter))

        with program.group("evaluator"):
            program.add_node(lp.CourierNode(self.evaluator, trainer, counter, trainer))

        if not self._num_caches:
            # Use the trainer as a single variable source.
            sources = [trainer]
        else:
            with program.group("cacher"):
                # Create a set of trainer caches.
                sources = []
                for _ in range(self._num_caches):
                    cacher = program.add_node(
                        lp.CacherNode(
                            trainer, refresh_interval_ms=2000, stale_after_ms=4000
                        )
                    )
                    sources.append(cacher)

        with program.group("executor"):
            # Add executors which pull round-robin from our variable sources.
            for executor_id in range(self._num_exectors):
                source = sources[executor_id % len(sources)]
                program.add_node(
                    lp.CourierNode(
                        self.executor,
                        executor_id,
                        replay,
                        source,
                        counter,
                        trainer,
                    )
                )

        return program



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Architectures using networked agents for multi-agent RL systems"""

import copy
from typing import Dict, List, Tuple

import sonnet as snt
import tensorflow as tf
from acme import specs as acme_specs

from mava import specs as mava_specs
from mava.components.tf.architectures.decentralised import (
    DecentralisedPolicyActor,
    DecentralisedQValueActorCritic,
)


class NetworkedPolicyActor(DecentralisedPolicyActor):
    """Networked multi-agent actor critic architecture."""

    def __init__(
        self,
        environment_spec: mava_specs.MAEnvironmentSpec,
        network_spec: Dict[str, List[str]],
        observation_networks: Dict[str, snt.Module],
        policy_networks: Dict[str, snt.Module],
        shared_weights: bool = False,
    ):
        super().__init__(
            environment_spec=environment_spec,
            observation_networks=observation_networks,
            policy_networks=policy_networks,
            shared_weights=shared_weights,
        )

        self._network_spec = network_spec

        if self._shared_weights:
            raise Exception(
                "Networked architectures currently do not support weight sharing."
            )

    def _get_actor_spec(self, agent_key: str) -> Dict[str, acme_specs.Array]:
        """Create network structure specifying connection between agents"""
        actor_obs_specs: Dict[str, acme_specs.Array] = {}

        agents_by_type = self._env_spec.get_agents_by_type()

        for agent_type, agents in agents_by_type.items():
            actor_obs_shape = list(
                copy.copy(
                    self._agent_type_specs[agent_type].observations.observation.shape
                )
            )
            for agent in agents:
                actor_obs_shape.insert(0, len(self._network_spec[agent]))
                actor_obs_specs[agent] = tf.TensorSpec(
                    shape=actor_obs_shape,
                    dtype=tf.dtypes.float32,
                )
        return actor_obs_specs


class NetworkedQValueCritic(DecentralisedQValueActorCritic):
    """Centralised multi-agent actor critic architecture."""

    def __init__(
        self,
        environment_spec: mava_specs.MAEnvironmentSpec,
        network_spec: Dict[str, List[str]],
        observation_networks: Dict[str, snt.Module],
        policy_networks: Dict[str, snt.Module],
        critic_networks: Dict[str, snt.Module],
        shared_weights: bool = True,
    ):
        super().__init__(
            environment_spec=environment_spec,
            observation_networks=observation_networks,
            policy_networks=policy_networks,
            critic_networks=critic_networks,
            shared_weights=shared_weights,
        )

        self._network_spec = network_spec

        if self._shared_weights:
            raise Exception(
                "Networked architectures currently do not support weight sharing."
            )

    def _get_critic_specs(
        self,
    ) -> Tuple[Dict[str, acme_specs.Array], Dict[str, acme_specs.Array]]:
        critic_obs_specs: Dict[str, acme_specs.Array] = {}
        critic_act_specs: Dict[str, acme_specs.Array] = {}

        agents_by_type = self._env_spec.get_agents_by_type()

        for agent_type, agents in agents_by_type.items():
            for agent in agents:
                critic_obs_shape = list(copy.copy(self._embed_specs[agent].shape))
                critic_act_shape = list(
                    copy.copy(self._agent_specs[agent].actions.shape)
                )
                critic_obs_shape.insert(0, len(self._network_spec[agent]))
                critic_obs_specs[agent] = tf.TensorSpec(
                    shape=critic_obs_shape,
                    dtype=tf.dtypes.float32,
                )
                critic_act_shape.insert(0, len(self._network_spec[agent]))
                critic_act_specs[agent] = tf.TensorSpec(
                    shape=critic_act_shape,
                    dtype=tf.dtypes.float32,
                )

        return critic_obs_specs, critic_act_specs



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wraps a PettingZoo MARL environment to be used as a dm_env environment."""
from typing import Any, Dict, Iterator, List, Optional, Union

import dm_env
import numpy as np
from acme import specs
from acme.wrappers.gym_wrapper import _convert_to_spec
from pettingzoo.utils.env import AECEnv, ParallelEnv
from supersuit import black_death_v1

from mava import types
from mava.utils.wrapper_utils import (
    apply_env_wrapper_preprocessers,
    convert_np_type,
    parameterized_restart,
)
from mava.wrappers.env_wrappers import ParallelEnvWrapper, SequentialEnvWrapper


# TODO(Kale-ab): Check usage agents vs possible agents
class PettingZooAECEnvWrapper(SequentialEnvWrapper):
    """Environment wrapper for PettingZoo MARL environments."""

    def __init__(
        self,
        environment: AECEnv,
        env_preprocess_wrappers: Optional[List] = [
            # (env_preprocessor, dict_with_preprocessor_params)
            (black_death_v1, None),
        ],
    ):
        self._environment = environment
        self._reset_next_step = True

        if env_preprocess_wrappers:
            self._environment = apply_env_wrapper_preprocessers(
                self._environment, env_preprocess_wrappers
            )

    def reset(self) -> dm_env.TimeStep:
        """Resets the episode."""
        self._reset_next_step = False
        self._environment.reset()
        self._step_type = dm_env.StepType.FIRST

        observe, _, done, _ = self._environment.last()
        agent = self.current_agent
        observation = self._convert_observation(agent, observe, done)

        self._discount = convert_np_type(
            self.discount_spec()[agent].dtype, 1
        )  # Not used in pettingzoo
        reward = convert_np_type(self.reward_spec()[agent].dtype, 0)
        return parameterized_restart(reward, self._discount, observation)

    def step(self, action: Union[int, float]) -> dm_env.TimeStep:
        """Steps the environment."""
        if self._reset_next_step:
            return self.reset()

        observe, reward, done, info = self._environment.last()

        # If current agent is done
        if done:
            self._environment.step(None)
            self._step_type = dm_env.StepType.LAST
        else:
            self._environment.step(action)
            self._step_type = dm_env.StepType.MID

            # Update these vars so dm_env.TimeStep returned has
            # the last information.
            observe, reward, done, info = self._environment.last()

        agent = self.current_agent
        # Convert rewards to match spec
        reward = convert_np_type(self.reward_spec()[agent].dtype, reward)
        observation = self._convert_observation(agent, observe, done)

        # Reset if all agents are done
        if self.env_done():
            self._reset_next_step = True

        return dm_env.TimeStep(
            observation=observation,
            reward=reward,
            discount=self._discount,
            step_type=self._step_type,
        )

    def env_done(self) -> bool:
        return self._environment.env_done

    def agent_iter(self, max_iter: int = 2 ** 63) -> Iterator:
        return self._environment.agent_iter(max_iter)

    # Convert PettingZoo observation so it's dm_env compatible. Also, the list
    # of legal actions must be converted to a legal actions mask.
    def _convert_observation(
        self, agent: str, observe: Union[dict, np.ndarray], done: bool
    ) -> types.OLT:
        if isinstance(observe, dict) and "action_mask" in observe:
            legals = observe["action_mask"]
            observe = observe["observation"]
        else:
            legals = np.ones(
                _convert_to_spec(self._environment.action_spaces[agent]).shape,
                dtype=self._environment.action_spaces[agent].dtype,
            )
        observation = types.OLT(
            observation=observe,
            legal_actions=legals,
            terminal=np.asarray([done], dtype=np.float32),
        )
        return observation

    def observation_spec(self) -> types.Observation:
        observation_specs = {}
        for agent in self.possible_agents:
            observation_specs[agent] = types.OLT(
                observation=_convert_to_spec(
                    self._environment.observation_spaces[agent]
                ),
                legal_actions=_convert_to_spec(self._environment.action_spaces[agent]),
                terminal=specs.Array((1,), np.float32),
            )
        return observation_specs

    def action_spec(self) -> Dict[str, specs.DiscreteArray]:
        action_specs = {}
        for agent in self.possible_agents:
            action_specs[agent] = _convert_to_spec(
                self._environment.action_spaces[agent]
            )
        return action_specs

    def reward_spec(self) -> Dict[str, specs.Array]:
        reward_specs = {}
        for agent in self.possible_agents:
            reward_specs[agent] = specs.Array((), np.float32)

        return reward_specs

    def discount_spec(self) -> Dict[str, specs.BoundedArray]:
        discount_specs = {}
        for agent in self.possible_agents:
            discount_specs[agent] = specs.BoundedArray(
                (), np.float32, minimum=0, maximum=1.0
            )
        return discount_specs

    def extra_spec(self) -> Dict[str, specs.BoundedArray]:
        return {}

    @property
    def agents(self) -> List:
        return self._environment.agents

    @property
    def possible_agents(self) -> List:
        return self._environment.possible_agents

    @property
    def environment(self) -> AECEnv:
        """Returns the wrapped environment."""
        return self._environment

    @property
    def current_agent(self) -> Any:
        return self._environment.agent_selection

    def __getattr__(self, name: str) -> Any:
        """Expose any other attributes of the underlying environment."""
        if hasattr(self.__class__, name):
            return self.__getattribute__(name)
        else:
            return getattr(self._environment, name)


class PettingZooParallelEnvWrapper(ParallelEnvWrapper):
    """Environment wrapper for PettingZoo MARL environments."""

    def __init__(
        self,
        environment: ParallelEnv,
        env_preprocess_wrappers: Optional[List] = [
            # (env_preprocessor, dict_with_preprocessor_params)
            (black_death_v1, None),
        ],
    ):
        self._environment = environment
        self._reset_next_step = True

        if env_preprocess_wrappers:
            self._environment = apply_env_wrapper_preprocessers(
                self._environment, env_preprocess_wrappers
            )

    def reset(self) -> dm_env.TimeStep:
        """Resets the episode."""
        self._reset_next_step = False
        self._step_type = dm_env.StepType.FIRST
        discount_spec = self.discount_spec()
        observe = self._environment.reset()

        self._discounts = {
            agent: convert_np_type(discount_spec[agent].dtype, 1)
            for agent in self.possible_agents
        }

        if type(observe) == tuple:
            observe, env_extras = observe
        else:
            env_extras = {}

        observations = self._convert_observations(
            observe, {agent: False for agent in self.possible_agents}
        )
        rewards_spec = self.reward_spec()
        rewards = {
            agent: convert_np_type(rewards_spec[agent].dtype, 0)
            for agent in self.possible_agents
        }

        return parameterized_restart(rewards, self._discounts, observations), env_extras

    def step(self, actions: Dict[str, np.ndarray]) -> dm_env.TimeStep:
        """Steps the environment."""

        if self._reset_next_step:
            return self.reset()

        observations, rewards, dones, infos = self._environment.step(actions)

        rewards_spec = self.reward_spec()
        #  Handle empty rewards
        if not rewards:
            rewards = {
                agent: convert_np_type(rewards_spec[agent].dtype, 0)
                for agent in self.possible_agents
            }
        else:
            rewards = {
                agent: convert_np_type(rewards_spec[agent].dtype, reward)
                for agent, reward in rewards.items()
            }

        if observations:
            observations = self._convert_observations(observations, dones)

        if self.env_done():
            self._step_type = dm_env.StepType.LAST
            self._reset_next_step = True
        else:
            self._step_type = dm_env.StepType.MID

        return dm_env.TimeStep(
            observation=observations,
            reward=rewards,
            discount=self._discounts,
            step_type=self._step_type,
        )

    def env_done(self) -> bool:
        return self._environment.env_done

    # Convert PettingZoo observation so it's dm_env compatible. Also, the list
    # of legal actions must be converted to a legal actions mask.
    def _convert_observations(
        self, observes: Dict[str, np.ndarray], dones: Dict[str, bool]
    ) -> types.Observation:
        observations: Dict[str, types.OLT] = {}
        for agent, observation in observes.items():
            if isinstance(observation, dict) and "action_mask" in observation:
                legals = observation["action_mask"]
                observation = observation["observation"]
            else:
                # TODO Handle legal actions better for continous envs,
                #  maybe have min and max for each action and clip the agents actions
                #  accordingly
                legals = np.ones(
                    _convert_to_spec(self._environment.action_spaces[agent]).shape,
                    dtype=self._environment.action_spaces[agent].dtype,
                )
            observations[agent] = types.OLT(
                observation=observation,
                legal_actions=legals,
                terminal=np.asarray([dones[agent]], dtype=np.float32),
            )

        return observations

    def observation_spec(self) -> types.Observation:
        observation_specs = {}
        for agent in self.possible_agents:
            observation_specs[agent] = types.OLT(
                observation=_convert_to_spec(
                    self._environment.observation_spaces[agent]
                ),
                legal_actions=_convert_to_spec(self._environment.action_spaces[agent]),
                terminal=specs.Array((1,), np.float32),
            )
        return observation_specs

    def action_spec(self) -> Dict[str, Union[specs.DiscreteArray, specs.BoundedArray]]:
        action_specs = {}
        action_spaces = self._environment.action_spaces
        for agent in self.possible_agents:
            action_specs[agent] = _convert_to_spec(action_spaces[agent])
        return action_specs

    def reward_spec(self) -> Dict[str, specs.Array]:
        reward_specs = {}
        for agent in self.possible_agents:
            reward_specs[agent] = specs.Array((), np.float32)

        return reward_specs

    def discount_spec(self) -> Dict[str, specs.BoundedArray]:
        discount_specs = {}
        for agent in self.possible_agents:
            discount_specs[agent] = specs.BoundedArray(
                (), np.float32, minimum=0, maximum=1.0
            )
        return discount_specs

    def extra_spec(self) -> Dict[str, specs.BoundedArray]:
        return {}

    @property
    def agents(self) -> List:
        return self._environment.agents

    @property
    def possible_agents(self) -> List:
        return self._environment.possible_agents

    @property
    def environment(self) -> ParallelEnv:
        """Returns the wrapped environment."""
        return self._environment

    @property
    def current_agent(self) -> Any:
        return self._environment.agent_selection

    def __getattr__(self, name: str) -> Any:
        """Expose any other attributes of the underlying environment."""
        if hasattr(self.__class__, name):
            return self.__getattribute__(name)
        else:
            return getattr(self._environment, name)



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc


class BaseExplorationScheduler:
    @abc.abstractmethod
    def __init__(
        self,
        epsilon_start: float = 1.0,
        epsilon_min: float = 0.05,
        epsilon_decay: float = 1e-4,
    ):
        """
        Base class for decaying epsilon by schedule.
        """
        self._epsilon_start = epsilon_start
        self._epsilon_min = epsilon_min
        self._epsilon_decay = epsilon_decay
        self._epsilon = epsilon_start

    @abc.abstractmethod
    def decrement_epsilon(self) -> None:
        """Decrement the epsilon decay value."""

    def get_epsilon(self) -> float:
        return self._epsilon

    def reset_epsilon(self) -> None:
        self._epsilon = self._epsilon_start


class LinearExplorationScheduler(BaseExplorationScheduler):
    def __init__(
        self,
        epsilon_start: float = 1.0,
        epsilon_min: float = 0.05,
        epsilon_decay: float = 1e-4,
    ):
        """
        Decays epsilon linearly to epsilon_min.
        """
        super(LinearExplorationScheduler, self).__init__(
            epsilon_start,
            epsilon_min,
            epsilon_decay,
        )

    def decrement_epsilon(self) -> None:
        if self._epsilon == self._epsilon_min:
            return

        elif self._epsilon < self._epsilon_min:
            # Should only ever happen once.
            self._epsilon = self._epsilon_min
            return

        self._epsilon -= self._epsilon_decay


class ExponentialExplorationScheduler(BaseExplorationScheduler):
    def __init__(
        self,
        epsilon_start: float = 1.0,
        epsilon_min: float = 0.05,
        epsilon_decay: float = 1e-4,
    ):
        """
        Decays epsilon exponentially to epsilon_min.
        """
        super(ExponentialExplorationScheduler, self).__init__(
            epsilon_start,
            epsilon_min,
            epsilon_decay,
        )

    def decrement_epsilon(self) -> None:
        if self._epsilon == self._epsilon_min:
            return

        elif self._epsilon < self._epsilon_min:
            # Should only ever happen once.
            self._epsilon = self._epsilon_min
            return

        self._epsilon *= 1 - self._epsilon_decay



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# TODO (Kevin): complete base class for communication

import abc
from typing import Dict, Union

import sonnet as snt

from mava.components.tf.architectures import BaseArchitecture, BasePolicyArchitecture

"""Base communication interface for multi-agent RL systems"""


class BaseCommunicationModule:
    """Base class for MARL communication.
    Objects which implement this interface provide a set of functions
    to create systems that can perform some form of communication between
    agents in a multi-agent RL system.
    """

    @abc.abstractmethod
    def __init__(
        self,
        architecture: Union[
            BaseArchitecture,
            BasePolicyArchitecture,
        ],
        shared: bool,
        channel_size: int,
        channel_noise: float,
    ) -> None:
        """Initializes the broadcaster communicator.
        Args:
            architecture: the BaseArchitecture used.
            shared: if a shared communication channel is used.
            channel_noise: stddev of normal noise in channel.
        """

    @abc.abstractmethod
    def process_messages(
        self, messages: Dict[str, snt.Module]
    ) -> Dict[str, snt.Module]:
        """Abstract communication function."""

    @abc.abstractmethod
    def create_system(self) -> Dict[str, Dict[str, snt.Module]]:
        """Create system architecture with communication."""
        # create_system on base architecture



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MARL system architectural components."""

from mava.components.tf.architectures.base import (
    BaseActorCritic,
    BaseArchitecture,
    BasePolicyArchitecture,
)
from mava.components.tf.architectures.centralised import (
    CentralisedPolicyActor,
    CentralisedQValueActorCritic,
    CentralisedQValueCritic,
    CentralisedValueCritic,
)
from mava.components.tf.architectures.decentralised import (
    DecentralisedPolicyActor,
    DecentralisedQValueActorCritic,
    DecentralisedValueActor,
    DecentralisedValueActorCritic,
)
from mava.components.tf.architectures.networked import (
    NetworkedPolicyActor,
    NetworkedQValueCritic,
)
from mava.components.tf.architectures.state_based import (
    StateBasedPolicyActor,
    StateBasedQValueActorCritic,
    StateBasedQValueCritic,
)



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Starcraft 2 environment factory."""

from typing import Any, Dict, Optional

import dm_env
from smac.env import StarCraft2Env

from mava.wrappers import SMACEnvWrapper  # type:ignore


def load_smac_env(env_config: Dict[str, Any]) -> StarCraft2Env:
    """Loads a smac environment given a config dict. Also, the possible agents in the
    environment are set"""

    env = StarCraft2Env(**env_config)
    env.possible_agents = list(range(env.n_agents))

    return env


def make_environment(
    evaluation: bool = False,
    map_name: str = "3m",
    random_seed: Optional[int] = None,
    **kwargs: Any,
) -> dm_env.Environment:
    """Wraps an starcraft 2 environment.

    Args:
        map_name: str, name of micromanagement level.

    Returns:
        A starcraft 2 smac environment wrapped as a DeepMind environment.
    """
    del evaluation

    env = StarCraft2Env(map_name=map_name, **kwargs)

    # wrap starcraft 2 environment
    environment = SMACEnvWrapper(env)

    if random_seed and hasattr(environment, "seed"):
        environment.seed(random_seed)

    return environment



# python3
# Copyright 2021 InstaDeep Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sonnet as snt
import tensorflow as tf


class AdditiveMixingNetwork(snt.Module):
    """Multi-agent monotonic mixing architecture."""

    def __init__(self) -> None:
        """Initializes the mixer."""
        super(AdditiveMixingNetwork, self).__init__()

    def __call__(self, q_values: tf.Tensor) -> tf.Tensor:
        """Monotonic mixing logic."""
        # return tf.math.reduce_sum(q_values, axis=1)
        return tf.math.reduce_sum(q_values, axis=1)
